from dataclasses import dataclass
import hashlib
from pathlib import Path
import uuid

import pytest
import requests
from pulpcore.exceptions.validation import InvalidSignatureError

from pulp_rpm.app.shared_utils import RpmTool
from pulp_rpm.tests.functional.constants import RPM_PACKAGE_FILENAME, RPM_UNSIGNED_URL
from pulp_rpm.tests.functional.utils import get_package_repo_path


def get_fixture(path: Path, url: str) -> Path:
    path.write_bytes(requests.get(url).content)
    return path


@pytest.mark.parallel
def test_register_rpm_package_signing_service(rpm_package_signing_service):
    """
    Register a sample rpmsign-based signing service and validate it works.
    """
    service = rpm_package_signing_service
    assert "/api/v3/signing-services/" in service.pulp_href


@dataclass
class GPGMetadata:
    pubkey: str
    fingerprint: str
    keyid: str


@pytest.fixture
def signing_gpg_extra(signing_gpg_metadata):
    """GPG instance with an extra gpg keypair registered."""
    PRIVATE_KEY_PULP_QE = (
        "https://raw.githubusercontent.com/pulp/pulp-fixtures/master/common/GPG-PRIVATE-KEY-pulp-qe"
    )
    gpg, fingerprint_a, keyid_a = signing_gpg_metadata

    response_private = requests.get(PRIVATE_KEY_PULP_QE)
    response_private.raise_for_status()
    import_result = gpg.import_keys(response_private.content)
    fingerprint_b = import_result.fingerprints[0]
    gpg.trust_keys(fingerprint_b, "TRUST_ULTIMATE")

    pubkey_a = gpg.export_keys(fingerprint_a)
    pubkey_b = gpg.export_keys(fingerprint_b)
    return (
        GPGMetadata(pubkey_a, fingerprint_a, fingerprint_a[-8:]),
        GPGMetadata(pubkey_b, fingerprint_b, fingerprint_b[-8:]),
    )


@pytest.mark.parallel
def test_sign_package_on_upload(
    tmp_path,
    pulpcore_bindings,
    monitor_task,
    gen_object_with_cleanup,
    download_content_unit,
    signing_gpg_extra,
    rpm_package_signing_service,
    rpm_package_api,
    rpm_repository_factory,
    rpm_publication_factory,
    rpm_package_factory,
    rpm_distribution_factory,
):
    """
    Sign an Rpm Package with the Package Upload endpoint.

    This ensures different
    """
    # Setup RPM tool and package to upload
    gpg_a, gpg_b = signing_gpg_extra
    fingerprint_set = set([gpg_a.fingerprint, gpg_b.fingerprint])
    assert len(fingerprint_set) == 2

    rpm_tool = RpmTool(tmp_path)
    rpm_tool.import_pubkey_string(gpg_a.pubkey)
    rpm_tool.import_pubkey_string(gpg_b.pubkey)

    file_to_upload = tmp_path / RPM_PACKAGE_FILENAME
    file_to_upload.write_bytes(requests.get(RPM_UNSIGNED_URL).content)
    with pytest.raises(InvalidSignatureError, match="The package is not signed: .*"):
        rpm_tool.verify_signature(file_to_upload)

    # Upload Package to Repository
    # The same file is uploaded, but signed with different keys each time
    for fingerprint in fingerprint_set:
        repository = rpm_repository_factory(
            package_signing_service=rpm_package_signing_service.pulp_href,
            package_signing_fingerprint=fingerprint,
        )
        upload_response = rpm_package_api.create(
            file=str(file_to_upload.absolute()),
            repository=repository.pulp_href,
        )
        package_href = monitor_task(upload_response.task).created_resources[2]
        pkg_location_href = rpm_package_api.read(package_href).location_href

        # Verify that the final served package is signed
        publication = rpm_publication_factory(repository=repository.pulp_href)
        distribution = rpm_distribution_factory(publication=publication.pulp_href)
        downloaded_package = tmp_path / "package.rpm"
        downloaded_package.write_bytes(
            download_content_unit(distribution.base_path, get_package_repo_path(pkg_location_href))
        )
        assert rpm_tool.verify_signature(downloaded_package)


@pytest.fixture
def pulpcore_chunked_file_factory(tmp_path):
    """Returns a function to create chunks from file to be uploaded."""

    def _create_chunks(upload_path, chunk_size=512):
        """Chunks file to be uploaded."""
        chunks = {"chunks": []}
        hasher = hashlib.new("sha256")
        start = 0
        with open(upload_path, "rb") as f:
            data = f.read()
        chunks["size"] = len(data)

        while start < len(data):
            content = data[start : start + chunk_size]
            chunk_file = tmp_path / str(uuid.uuid4())
            hasher.update(content)
            chunk_file.write_bytes(content)
            content_sha = hashlib.sha256(content).hexdigest()
            end = start + len(content) - 1
            chunks["chunks"].append(
                (str(chunk_file), f"bytes {start}-{end}/{chunks['size']}", content_sha)
            )
            start += len(content)
        chunks["digest"] = hasher.hexdigest()
        return chunks

    return _create_chunks


@pytest.fixture
def pulpcore_upload_chunks(
    pulpcore_bindings,
    gen_object_with_cleanup,
    monitor_task,
):
    """Upload file in chunks."""

    def _upload_chunks(size, chunks, sha256, include_chunk_sha256=False):
        """
        Chunks is a list of tuples in the form of (chunk_filename, "bytes-ranges", optional_sha256).
        """
        upload = gen_object_with_cleanup(pulpcore_bindings.UploadsApi, {"size": size})

        for data in chunks:
            kwargs = {"file": data[0], "content_range": data[1], "upload_href": upload.pulp_href}
            if include_chunk_sha256:
                if len(data) != 3:
                    raise Exception(f"Chunk didn't include its sha256: {data}")
                kwargs["sha256"] = data[2]

            pulpcore_bindings.UploadsApi.update(**kwargs)

        return upload

    yield _upload_chunks


@pytest.mark.parallel
def test_sign_chunked_package_on_upload(
    tmp_path,
    pulpcore_bindings,
    monitor_task,
    gen_object_with_cleanup,
    download_content_unit,
    signing_gpg_extra,
    rpm_package_signing_service,
    rpm_package_api,
    rpm_repository_factory,
    rpm_publication_factory,
    rpm_package_factory,
    rpm_distribution_factory,
    pulpcore_upload_chunks,
    pulpcore_chunked_file_factory,
):
    """
    Sign an Rpm Package with the Package Upload endpoint.

    This ensures different
    """
    # Setup RPM tool and package to upload
    gpg_a, gpg_b = signing_gpg_extra
    fingerprint_set = set([gpg_a.fingerprint, gpg_b.fingerprint])
    assert len(fingerprint_set) == 2

    rpm_tool = RpmTool(tmp_path)
    rpm_tool.import_pubkey_string(gpg_a.pubkey)
    rpm_tool.import_pubkey_string(gpg_b.pubkey)

    file_to_upload = tmp_path / RPM_PACKAGE_FILENAME
    file_to_upload.write_bytes(requests.get(RPM_UNSIGNED_URL).content)
    with pytest.raises(InvalidSignatureError, match="The package is not signed: .*"):
        rpm_tool.verify_signature(file_to_upload)

    # Upload Package to Repository
    # The same file is uploaded, but signed with different keys each time
    for fingerprint in fingerprint_set:
        repository = rpm_repository_factory(
            package_signing_service=rpm_package_signing_service.pulp_href,
            package_signing_fingerprint=fingerprint,
        )
        file_chunks_data = pulpcore_chunked_file_factory(file_to_upload)
        size = file_chunks_data["size"]
        chunks = file_chunks_data["chunks"]
        sha256 = file_chunks_data["digest"]

        upload = pulpcore_upload_chunks(size, chunks, sha256, include_chunk_sha256=True)
        upload_response = rpm_package_api.create(
            upload=upload.pulp_href,
            repository=repository.pulp_href,
        )
        package_href = monitor_task(upload_response.task).created_resources[2]
        pkg_location_href = rpm_package_api.read(package_href).location_href

        # Verify that the final served package is signed
        publication = rpm_publication_factory(repository=repository.pulp_href)
        distribution = rpm_distribution_factory(publication=publication.pulp_href)
        downloaded_package = tmp_path / "package.rpm"
        downloaded_package.write_bytes(
            download_content_unit(distribution.base_path, get_package_repo_path(pkg_location_href))
        )
        assert rpm_tool.verify_signature(downloaded_package)



from gettext import gettext as _

from pulpcore.plugin.exceptions import PulpException


class AdvisoryConflict(PulpException):
    """
    Raised when two advisories conflict in a way that Pulp can't resolve it.
    """

    def __init__(self, msg):
        """
        Set the exception identifier.

        Args:
            msg(str): Detailed message about the reasons for Advisory conflict
        """
        super().__init__("RPM0001")
        self.msg = msg

    def __str__(self):
        """
        Return a message for the exception.
        """
        return self.msg


class DistributionTreeConflict(FileNotFoundError):
    """
    Raised when two or more distribution trees are being added to a repository version.
    """

    def __init__(self, msg):
        """
        Set the exception identifier and msg.
        """
        super().__init__("RPM0002")
        self.msg = _("More than one distribution tree cannot be added to a " "repository version.")

    def __str__(self):
        """
        Return a message for the exception.
        """
        return self.msg


class UlnCredentialsError(PulpException):
    """
    Raised when no valid ULN Credentials were given.
    """

    def __init__(self, msg):
        """
        Set the exception identifier and msg.
        """
        super().__init__("RPM0003")
        self.msg = _("No valid ULN credentials given.")

    def __str__(self):
        """
        Return a message for the exception.
        """
        return self.msg



"""Tests that sync rpm plugin repositories."""

import pytest
import os
from datetime import datetime

from pulp_rpm.tests.functional.constants import (
    PULP_TYPE_REPOMETADATA,
    RHEL8_APPSTREAM_CDN_URL,
    RHEL8_BASEOS_CDN_URL,
    RPM_KICKSTART_CONTENT_NAME,
    RPM_KICKSTART_COUNT,
    CENTOS8_STREAM_APPSTREAM_URL,
    CENTOS8_STREAM_BASEOS_URL,
    CENTOS10_STREAM_APPSTREAM_URL,
    CENTOS10_STREAM_BASEOS_URL,
    EPEL8_MIRRORLIST_URL,
    RAWHIDE_KICKSTART_URL,
)


def parse_date_from_string(s, parse_format="%Y-%m-%dT%H:%M:%S.%fZ"):
    """Parse string to datetime object.

    :param s: str like '2018-11-18T21:03:32.493697Z'
    :param parse_format: str defaults to %Y-%m-%dT%H:%M:%S.%fZ
    :return: datetime.datetime
    """
    if isinstance(s, datetime):
        return s
    else:
        return datetime.strptime(s, parse_format)


@pytest.mark.parametrize(
    "url,policy,check_dist_tree,resync",
    [
        pytest.param(
            CENTOS8_STREAM_BASEOS_URL,
            "immediate",
            True,
            True,
            marks=pytest.mark.skip("Skip to avoid failing due to running out of disk space"),
        ),
        (CENTOS8_STREAM_BASEOS_URL, "on_demand", True, True),
        (CENTOS8_STREAM_APPSTREAM_URL, "on_demand", True, True),
        (EPEL8_MIRRORLIST_URL, "on_demand", False, False),
        (RAWHIDE_KICKSTART_URL, "on_demand", True, True),
        (CENTOS10_STREAM_BASEOS_URL, "on_demand", True, True),
        (CENTOS10_STREAM_APPSTREAM_URL, "on_demand", True, True),
    ],
)
def test_rpm_sync(
    url,
    policy,
    check_dist_tree,
    resync,
    init_and_sync,
    rpm_repository_version_api,
    delete_orphans_pre,
):
    """Sync repositories with the rpm plugin."""
    # Create repository & remote and sync
    repo, remote, task = init_and_sync(url=url, policy=policy, return_task=True)

    created_at = parse_date_from_string(task.pulp_created)
    started_at = parse_date_from_string(task.started_at)
    finished_at = parse_date_from_string(task.finished_at)
    task_duration = finished_at - started_at
    waiting_time = started_at - created_at
    print(
        "\n->     Sync => Waiting time (s): {wait} | Service time (s): {service}".format(
            wait=waiting_time.total_seconds(), service=task_duration.total_seconds()
        )
    )

    # Check that we have the correct content counts.
    if check_dist_tree:
        repo_ver = rpm_repository_version_api.read(repo.latest_version_href)
        present_summary = repo_ver.content_summary.present
        assert RPM_KICKSTART_CONTENT_NAME in present_summary
        assert present_summary[RPM_KICKSTART_CONTENT_NAME]["count"] == RPM_KICKSTART_COUNT
        added_summary = repo_ver.content_summary.added
        assert RPM_KICKSTART_CONTENT_NAME in added_summary
        assert added_summary[RPM_KICKSTART_CONTENT_NAME]["count"] == RPM_KICKSTART_COUNT

    if resync:
        # Sync the repository again.
        latest_version_href = repo.latest_version_href
        repo, remote, task = init_and_sync(repository=repo, remote=remote, return_task=True)
        created_at = parse_date_from_string(task.pulp_created)
        started_at = parse_date_from_string(task.started_at)
        finished_at = parse_date_from_string(task.finished_at)
        task_duration = finished_at - started_at
        waiting_time = started_at - created_at
        print(
            "\n->  Re-sync => Waiting time (s): {wait} | Service time (s): {service}".format(
                wait=waiting_time.total_seconds(), service=task_duration.total_seconds()
            )
        )

        # Check that nothing has changed since the last sync.
        assert latest_version_href == repo.latest_version_href


@pytest.fixture
def cdn_certs_and_keys():
    cdn_client_cert = os.getenv("CDN_CLIENT_CERT", "").replace("\\n", "\n")
    cdn_client_key = os.getenv("CDN_CLIENT_KEY", "").replace("\\n", "\n")
    cdn_ca_cert = os.getenv("CDN_CA_CERT", "").replace("\\n", "\n")

    return cdn_client_cert, cdn_client_key, cdn_ca_cert


def test_sync_with_certificate(
    cdn_certs_and_keys,
    init_and_sync,
    rpm_rpmremote_factory,
    rpm_content_repometadata_files_api,
    rpm_repository_version_api,
    delete_orphans_pre,
):
    """Test sync against CDN.

    1. create repository, appstream remote and sync
        - remote using certificates and tls validation
    2. create repository, baseos remote and sync
        - remote using certificates without tls validation
    3. Check both repositories were synced and both have its own 'productid' content
        - this test covering checking same repo metadata files with different relative paths
    """
    client_cert, client_key, ca_cert = cdn_certs_and_keys
    if not all(cdn_certs_and_keys):
        pytest.skip("CDN Client Cert & Key and CA Cert were not set")

    # 1. create repo, remote and sync them
    appstream_remote = rpm_rpmremote_factory(
        url=RHEL8_APPSTREAM_CDN_URL,
        client_cert=client_cert,
        client_key=client_key,
        ca_cert=ca_cert,
        policy="on_demand",
    )
    repo_appstream, _ = init_and_sync(remote=appstream_remote)

    # 2. create remote and re-sync
    baseos_remote = rpm_rpmremote_factory(
        url=RHEL8_BASEOS_CDN_URL,
        tls_validation=False,
        client_cert=client_cert,
        client_key=client_key,
        policy="on_demand",
    )
    repo_baseos, _ = init_and_sync(remote=baseos_remote)

    # Get all 'productid' repo metadata files
    productids = [
        productid
        for productid in rpm_content_repometadata_files_api.list().results
        if productid.data_type == "productid"
    ]

    # Assert there are two productid content units and they have same checksum
    assert len(productids) == 2
    assert productids[0].checksum == productids[1].checksum

    # Assert each repository has its own productid file
    appstream_repo_ver = rpm_repository_version_api.read(repo_appstream.latest_version_href)
    assert appstream_repo_ver.content_summary.present[PULP_TYPE_REPOMETADATA]["count"] == 1
    baseos_repo_ver = rpm_repository_version_api.read(repo_baseos.latest_version_href)
    assert baseos_repo_ver.content_summary.present[PULP_TYPE_REPOMETADATA]["count"] == 1

    appstream_metadata = rpm_content_repometadata_files_api.list(
        repository_version=appstream_repo_ver.pulp_href
    )
    baseos_metadata = rpm_content_repometadata_files_api.list(
        repository_version=baseos_repo_ver.pulp_href
    )
    assert appstream_metadata.results[0].relative_path != baseos_metadata.results[0].relative_path



# Generated by Django 4.2.10 on 2025-05-09 05:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0063_rpmpublication_layout_rpmrepository_layout'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='rpmrepository',
            name='original_checksum_types',
        ),
        migrations.AlterField(
            model_name='rpmpublication',
            name='metadata_checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], null=True),
        ),
        migrations.AlterField(
            model_name='rpmpublication',
            name='package_checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], null=True),
        ),
    ]



"""Tests that perform actions over content unit."""

from textwrap import dedent

import pytest
from pulpcore.client.pulp_rpm.exceptions import ForbiddenException
from pulpcore.client.pulp_rpm import RpmModulemdDefaults, RpmModulemd, SetLabel, UnsetLabel

from pulp_rpm.tests.functional.constants import (
    RPM_ADVISORY_CONTENT_NAME,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_MODULAR_FIXTURE_URL,
    RPM_MODULAR_DEFAULTS_CONTENT_NAME,
    RPM_MODULAR_MODULES_CONTENT_NAME,
    RPM_MODULES_OBSOLETE_CONTENT_NAME,
    RPM_PACKAGE_CONTENT_NAME,
    RPM_PACKAGE_FILENAME,
    RPM_PACKAGE_FILENAME2,
    RPM_REPO_METADATA_FIXTURE_URL,
)
from pulp_rpm.tests.functional.utils import gen_rpm_content_attrs
from pulpcore.tests.functional.utils import PulpTaskError


def test_crud_content_unit(
    delete_orphans_pre,
    signed_artifact,
    gen_object_with_cleanup,
    rpm_package_api,
    rpm_repository_api,
    rpm_repository_factory,
    rpm_repository_version_api,
    monitor_task,
):
    """Test creating, reading, updating, and deleting a content unit of package type."""
    # Create content unit

    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME)
    response = rpm_package_api.create(**attrs)
    content_unit = rpm_package_api.read(monitor_task(response.task).created_resources[0])
    # rpm package doesn't keep relative_path but the location href
    del attrs["relative_path"]

    for key, val in attrs.items():
        assert getattr(content_unit, key) == val

    # Read a content unit by its href
    response = rpm_package_api.read(content_unit.pulp_href)
    assert response == content_unit

    # Read a content unit by its pkg_id
    page = rpm_package_api.list(pkg_id=content_unit.pkg_id)
    assert len(page.results) == 1
    assert page.results[0] == content_unit

    # Attempt to update a content unit using HTTP PATCH
    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME2)
    with pytest.raises(AttributeError) as exc:
        rpm_package_api.partial_update(content_unit.pulp_href, attrs)
    msg = "object has no attribute 'partial_update'"
    assert msg in str(exc)

    # Attempt to update a content unit using HTTP PUT
    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME2)
    with pytest.raises(AttributeError) as exc:
        rpm_package_api.update(content_unit.pulp_href, attrs)
    msg = "object has no attribute 'update'"
    assert msg in str(exc)

    # Attempt to delete a content unit using HTTP DELETE
    with pytest.raises(AttributeError) as exc:
        rpm_package_api.delete(content_unit.pulp_href)
    msg = "object has no attribute 'delete'"
    assert msg in str(exc)

    # Attempt to create duplicate package without specifying a repository
    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME)
    response = rpm_package_api.create(**attrs)
    duplicate = rpm_package_api.read(monitor_task(response.task).created_resources[0])
    assert duplicate.pulp_href == content_unit.pulp_href

    # Attempt to create duplicate package while specifying a repository
    repo = rpm_repository_factory()
    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME)
    attrs["repository"] = repo.pulp_href
    response = rpm_package_api.create(**attrs)
    monitored_response = monitor_task(response.task)

    duplicate = rpm_package_api.read(monitored_response.created_resources[1])
    assert duplicate.pulp_href == content_unit.pulp_href

    repo = rpm_repository_api.read(repo.pulp_href)
    assert repo.latest_version_href.endswith("/versions/1/")

    version = rpm_repository_version_api.read(repo.latest_version_href)
    assert version.content_summary.present["rpm.package"]["count"] == 1
    assert version.content_summary.added["rpm.package"]["count"] == 1


@pytest.mark.parallel
def test_set_unset_search_labels(
    gen_user,
    get_content,
    init_and_sync,
    rpm_advisory_api,
    rpm_modulemd_api,
    rpm_modulemd_defaults_api,
    rpm_modulemd_obsoletes_api,
    rpm_package_api,
):
    """
    Sync a repository and test the set/unset/search label calls for
    content-types which support that.
    """

    def _do_test(binding, content_type_name, repo_content, rv):
        a_content = binding.read(repo_content["present"][content_type_name][0]["pulp_href"])

        # Set label
        sl = SetLabel(key="key_1", value="value_1")
        binding.set_label(a_content.pulp_href, sl)
        a_content = binding.read(a_content.pulp_href)
        labels = a_content.pulp_labels
        assert "key_1" in labels

        # Search for key_1
        rslt = binding.list(pulp_label_select="key_1", repository_version=rv)
        assert 1 == rslt.count

        # Change an existing label
        sl = SetLabel(key="key_1", value="XXX")
        binding.set_label(a_content.pulp_href, sl)
        a_content = binding.read(a_content.pulp_href)
        labels = a_content.pulp_labels
        assert labels["key_1"] == "XXX"

        # Unset a label
        sl = UnsetLabel(key="key_1")
        binding.unset_label(a_content.pulp_href, sl)
        content2 = binding.read(a_content.pulp_href)
        assert "key_1" not in content2.pulp_labels

    # Set up and sync a repository with the desired content-types
    # Show that content-labeler can set/unset_label
    repo_owner = gen_user(model_roles=["rpm.admin", "core.content_labeler"])
    with repo_owner:
        repository, _ = init_and_sync(url=RPM_MODULAR_FIXTURE_URL)
        repository_content = get_content(repository)

        # Test set/unset/search for each type-of-content
        # We don't do this via pytest-parameterization so that we only sync the repo *once*.
        content_type_tuples = [
            (rpm_package_api, RPM_PACKAGE_CONTENT_NAME),
            (rpm_advisory_api, RPM_ADVISORY_CONTENT_NAME),
            (rpm_modulemd_api, RPM_MODULAR_MODULES_CONTENT_NAME),
            (rpm_modulemd_defaults_api, RPM_MODULAR_DEFAULTS_CONTENT_NAME),
            (rpm_modulemd_obsoletes_api, RPM_MODULES_OBSOLETE_CONTENT_NAME),
        ]
        for t in content_type_tuples:
            _do_test(t[0], t[1], repository_content, repository.latest_version_href)

    # Show that a repository-viewer DOES NOT HAVE access to set/unset_label
    viewer = gen_user(model_roles=["rpm.rpmrepository_viewer"])
    with viewer:
        # repository, _ = init_and_sync(url=RPM_MODULAR_FIXTURE_URL)
        # repo_content = get_content(repository)

        # Test set/unset/search for each type-of-content
        # We don't do this via pytest-parameterization so that we only sync the repo *once*.
        content_type_tuples = [
            (rpm_package_api, RPM_PACKAGE_CONTENT_NAME),
            (rpm_advisory_api, RPM_ADVISORY_CONTENT_NAME),
            (rpm_modulemd_api, RPM_MODULAR_MODULES_CONTENT_NAME),
            (rpm_modulemd_defaults_api, RPM_MODULAR_DEFAULTS_CONTENT_NAME),
            (rpm_modulemd_obsoletes_api, RPM_MODULES_OBSOLETE_CONTENT_NAME),
        ]
        for t in content_type_tuples:
            content = t[0].read(repository_content["present"][t[1]][0]["pulp_href"])

            with pytest.raises(ForbiddenException):
                label = SetLabel(key="key_1", value="XXX")
                t[0].set_label(content.pulp_href, label)

            with pytest.raises(ForbiddenException):
                label = UnsetLabel(key="key_1")
                t[0].unset_label(content.pulp_href, label)


@pytest.mark.parallel
@pytest.mark.parametrize(
    "url",
    [RPM_MODULAR_FIXTURE_URL, RPM_KICKSTART_FIXTURE_URL, RPM_REPO_METADATA_FIXTURE_URL],
    ids=["MODULAR_FIXTURE_URL", "KICKSTART_FIXTURE_URL", "REPO_METADATA_FIXTURE_URL"],
)
def test_remove_content_unit(url, init_and_sync, get_content, pulp_requests):
    """
    Sync a repository and test that content of any type cannot be removed directly.

    - advisory
    - distribution_tree
    - modulemd
    - modulemd_defaults
    - package
    - packagecategory
    - packageenvironment
    - packagegroup
    - packagelangpacks
    - repo metadata
    """
    # Test remove content by types contained in repository.
    repo, _ = init_and_sync(url=url, policy="on_demand")
    added_content = get_content(repo)["added"]

    # iterate over content units and issue delete requests
    for content_type, content_list in added_content.items():
        for content_unit in content_list:
            resp = pulp_requests.delete(content_unit["pulp_href"])
            assert resp.status_code == 405  # method not allowed


def test_create_modulemd_defaults(monitor_task, gen_object_with_cleanup, rpm_modulemd_defaults_api):
    """
    Create modulemd_defaults with proper unique identifier.

    See: https://github.com/pulp/pulp_rpm/issues/3495
    """
    request_1 = {
        "module": "squid",
        "stream": "4",
        "profiles": {"4": ["common"]},
        "snippet": dedent(
            """\
        ---
        document: modulemd-defaults
        version: 1
        data:
          module: squid
          stream: "4"
          profiles:
            4: [common]
        ..."""
        ),
    }

    # Can create
    modulemd_default = gen_object_with_cleanup(
        rpm_modulemd_defaults_api, RpmModulemdDefaults(**request_1)
    )
    assert modulemd_default.module == request_1["module"]

    # Can create variation
    request_2 = request_1.copy()
    request_2["snippet"] = request_2["snippet"].replace("module: squid", 'module: "squid-mod"')
    modulemd_default = gen_object_with_cleanup(
        rpm_modulemd_defaults_api, RpmModulemdDefaults(**request_2)
    )
    assert modulemd_default.module == request_2["module"]

    # Cant create duplicate
    request_3 = request_1.copy()
    request_3["module"] = "squid-mod2"  # not in unique_togheter
    with pytest.raises(PulpTaskError) as exc:
        modulemd_default = gen_object_with_cleanup(
            rpm_modulemd_defaults_api, RpmModulemdDefaults(**request_3)
        )
    assert "duplicate key value violates unique constraint" in exc.value.task.error["description"]


def test_create_modulemds(
    monitor_task, gen_object_with_cleanup, rpm_modulemd_api, rpm_package_factory
):
    package = rpm_package_factory()
    request = {
        "name": "foo",
        "stream": "foo",
        "version": "foo",
        "context": "foo",
        "arch": "foo",
        "artifacts": "[]",
        "dependencies": "[]",
        "packages": [package.pulp_href],
        "snippet": "foobar",
        "profiles": "[]",
        "description": "foo",
    }

    # Can upload
    modulemd = gen_object_with_cleanup(rpm_modulemd_api, RpmModulemd(**request))
    assert modulemd.name == request["name"]

    # Cant create duplicate
    with pytest.raises(PulpTaskError) as exc:
        modulemd = gen_object_with_cleanup(rpm_modulemd_api, RpmModulemd(**request))
    assert "duplicate key value violates unique constraint" in exc.value.task.error["description"]

    # Can upload variation
    request2 = request.copy()
    request2["snippet"] = "barfoo"
    modulemd = gen_object_with_cleanup(rpm_modulemd_api, RpmModulemd(**request2))
    assert modulemd.name == request2["name"]



# Generated by Django 4.2.10 on 2025-07-01 20:29

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0064_remove_rpmrepository_original_checksum_types_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='package',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('upload_rpm_packages', 'Can upload RPM packages using synchronous API.')]},
        ),
    ]



"""Tests for checkpoint distribution and publications."""

from datetime import datetime, timedelta
import re
from time import sleep
from urllib.parse import urlparse
from aiohttp import ClientResponseError
from tempfile import NamedTemporaryFile
import pytest
import requests
from pulp_rpm.tests.functional.constants import RPM_SIGNED_URL


@pytest.fixture(scope="class")
def rpm_package_factory_class(
    gen_object_with_cleanup,
    pulp_domain_enabled,
    rpm_package_api,
):
    """Return a Package created from uploading an RPM file."""

    def _rpm_package_factory_class(url=RPM_SIGNED_URL, pulp_domain=None):
        with NamedTemporaryFile() as file_to_upload:
            file_to_upload.write(requests.get(url).content)
            file_to_upload.flush()
            upload_attrs = {"file": file_to_upload.name}

            kwargs = {}
            if pulp_domain:
                if not pulp_domain_enabled:
                    raise RuntimeError("Server does not have domains enabled.")
                kwargs["pulp_domain"] = pulp_domain

            return gen_object_with_cleanup(rpm_package_api, **upload_attrs, **kwargs)

    return _rpm_package_factory_class


@pytest.fixture(scope="class")
def setup(
    rpm_repository_factory,
    rpm_publication_factory,
    rpm_distribution_factory,
    rpm_package_factory_class,
    rpm_repository_api,
    monitor_task,
):
    def create_publication(repo, checkpoint):
        artifact = rpm_package_factory_class()
        monitor_task(
            rpm_repository_api.modify(
                repo.pulp_href, {"add_content_units": [artifact.pulp_href]}
            ).task
        )
        return rpm_publication_factory(repository=repo.pulp_href, checkpoint=checkpoint)

    repo = rpm_repository_factory()
    distribution = rpm_distribution_factory(repository=repo.pulp_href, checkpoint=True)

    pubs = []
    # Create publications with 1-second intervals to ensure different timestamps/checkpoint URLs
    pubs.append(create_publication(repo, False))
    sleep(1)
    pubs.append(create_publication(repo, True))
    sleep(1)
    pubs.append(create_publication(repo, False))
    sleep(1)
    pubs.append(create_publication(repo, True))
    sleep(1)
    pubs.append(create_publication(repo, False))

    return pubs, distribution


@pytest.fixture
def checkpoint_url(distribution_base_url):
    def _checkpoint_url(distribution, timestamp):
        distro_base_url = distribution_base_url(distribution.base_url)
        return f"{distro_base_url}{_format_checkpoint_timestamp(timestamp)}/"

    return _checkpoint_url


def _format_checkpoint_timestamp(timestamp):
    return datetime.strftime(timestamp, "%Y%m%dT%H%M%SZ")


class TestCheckpointDistribution:
    @pytest.mark.parallel
    def test_base_path_lists_checkpoints(self, setup, http_get, distribution_base_url):
        pubs, distribution = setup

        response = http_get(distribution_base_url(distribution.base_url)).decode("utf-8")

        checkpoints_ts = set(re.findall(r"\d{8}T\d{6}Z", response))
        assert len(checkpoints_ts) == 2
        assert _format_checkpoint_timestamp(pubs[1].pulp_created) in checkpoints_ts
        assert _format_checkpoint_timestamp(pubs[3].pulp_created) in checkpoints_ts

    @pytest.mark.parallel
    def test_no_trailing_slash_is_redirected(self, setup, http_get, distribution_base_url):
        """Test checkpoint listing when path doesn't end with a slash."""

        pubs, distribution = setup

        response = http_get(distribution_base_url(distribution.base_url[:-1])).decode("utf-8")
        checkpoints_ts = set(re.findall(r"\d{8}T\d{6}Z", response))

        assert len(checkpoints_ts) == 2
        assert _format_checkpoint_timestamp(pubs[1].pulp_created) in checkpoints_ts
        assert _format_checkpoint_timestamp(pubs[3].pulp_created) in checkpoints_ts

    @pytest.mark.parallel
    def test_exact_timestamp_is_served(self, setup, http_get, checkpoint_url):
        pubs, distribution = setup

        pub_1_url = checkpoint_url(distribution, pubs[1].pulp_created)
        response = http_get(pub_1_url).decode("utf-8")

        assert f"<h1>Index of {urlparse(pub_1_url).path}</h1>" in response

    @pytest.mark.parallel
    def test_invalid_timestamp_returns_404(self, setup, http_get, distribution_base_url):
        _, distribution = setup
        with pytest.raises(ClientResponseError) as exc:
            http_get(distribution_base_url(f"{distribution.base_url}invalid_ts/"))

        assert exc.value.status == 404

        with pytest.raises(ClientResponseError) as exc:
            http_get(distribution_base_url(f"{distribution.base_url}20259928T092752Z/"))

        assert exc.value.status == 404

    @pytest.mark.parallel
    def test_non_checkpoint_timestamp_is_redirected(self, setup, http_get, checkpoint_url):
        pubs, distribution = setup
        # Using a non-checkpoint publication timestamp
        pub_3_url = checkpoint_url(distribution, pubs[3].pulp_created)
        pub_4_url = checkpoint_url(distribution, pubs[4].pulp_created)

        response = http_get(pub_4_url).decode("utf-8")
        assert f"<h1>Index of {urlparse(pub_3_url).path}</h1>" in response

        # Test without a trailing slash
        response = http_get(pub_4_url[:-1]).decode("utf-8")
        assert f"<h1>Index of {urlparse(pub_3_url).path}</h1>" in response

    @pytest.mark.parallel
    def test_arbitrary_timestamp_is_redirected(self, setup, http_get, checkpoint_url):
        pubs, distribution = setup
        pub_1_url = checkpoint_url(distribution, pubs[1].pulp_created)
        arbitrary_url = checkpoint_url(distribution, pubs[1].pulp_created + timedelta(seconds=1))

        response = http_get(arbitrary_url).decode("utf-8")
        assert f"<h1>Index of {urlparse(pub_1_url).path}</h1>" in response

        # Test without a trailing slash
        response = http_get(arbitrary_url[:-1]).decode("utf-8")
        assert f"<h1>Index of {urlparse(pub_1_url).path}</h1>" in response

    @pytest.mark.parallel
    def test_current_timestamp_serves_latest_checkpoint(self, setup, http_get, checkpoint_url):
        pubs, distribution = setup
        pub_3_url = checkpoint_url(distribution, pubs[3].pulp_created)
        now_url = checkpoint_url(distribution, datetime.now())

        response = http_get(now_url).decode("utf-8")

        assert f"<h1>Index of {urlparse(pub_3_url).path}</h1>" in response

    @pytest.mark.parallel
    def test_before_first_timestamp_returns_404(self, setup, http_get, checkpoint_url):
        pubs, distribution = setup
        pub_0_url = checkpoint_url(distribution, pubs[0].pulp_created)

        with pytest.raises(ClientResponseError) as exc:
            http_get(pub_0_url).decode("utf-8")

        assert exc.value.status == 404

    @pytest.mark.parallel
    def test_future_timestamp_returns_404(self, setup, http_get, checkpoint_url):
        _, distribution = setup
        url = checkpoint_url(distribution, datetime.now() + timedelta(days=1))

        with pytest.raises(ClientResponseError) as exc:
            http_get(url).decode("utf-8")

        assert exc.value.status == 404



# Generated by Django 3.2.13 on 2022-05-17 20:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0042_alter_repometadatafile_data_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='addon',
            name='addon_id',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='addon',
            name='name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='addon',
            name='packages',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='addon',
            name='type',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='addon',
            name='uid',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='checksum',
            name='checksum',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='checksum',
            name='path',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='arch',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='base_product_name',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='base_product_short',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='base_product_version',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='digest',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='header_version',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='instimage',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='mainimage',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='release_name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='release_short',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='distributiontree',
            name='release_version',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='image',
            name='name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='image',
            name='path',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='image',
            name='platforms',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='modulemd',
            name='arch',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='modulemd',
            name='context',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='modulemd',
            name='name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='modulemd',
            name='stream',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='modulemd',
            name='version',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='modulemddefaults',
            name='digest',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='modulemddefaults',
            name='module',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='modulemddefaults',
            name='stream',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='package',
            name='checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')]),
        ),
        migrations.AlterField(
            model_name='package',
            name='pkgId',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='packagecategory',
            name='digest',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='packagecategory',
            name='id',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='packagecategory',
            name='name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='packageenvironment',
            name='digest',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='packageenvironment',
            name='id',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='packageenvironment',
            name='name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='packagegroup',
            name='digest',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='packagegroup',
            name='id',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='packagegroup',
            name='name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='packagelangpacks',
            name='digest',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='repometadatafile',
            name='checksum',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='repometadatafile',
            name='checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')]),
        ),
        migrations.AlterField(
            model_name='rpmpublication',
            name='metadata_checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')]),
        ),
        migrations.AlterField(
            model_name='rpmpublication',
            name='package_checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')]),
        ),
        migrations.AlterField(
            model_name='rpmremote',
            name='sles_auth_token',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='rpmrepository',
            name='metadata_checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], null=True),
        ),
        migrations.AlterField(
            model_name='rpmrepository',
            name='package_checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], null=True),
        ),
        migrations.AlterField(
            model_name='ulnremote',
            name='uln_server_base_url',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='updaterecord',
            name='digest',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='updaterecord',
            name='id',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterField(
            model_name='variant',
            name='debug_packages',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='variant',
            name='debug_repository',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='variant',
            name='identity',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='variant',
            name='name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='variant',
            name='packages',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='variant',
            name='source_packages',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='variant',
            name='source_repository',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='variant',
            name='type',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='variant',
            name='uid',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='variant',
            name='variant_id',
            field=models.TextField(),
        ),
    ]



# Generated by Django 4.2.10 on 2025-02-12 19:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0062_rpmpackagesigningservice_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmpublication',
            name='layout',
            field=models.TextField(choices=[('nested_alphabetically', 'nested_alphabetically'), ('flat', 'flat')], null=True),
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='layout',
            field=models.TextField(choices=[('nested_alphabetically', 'nested_alphabetically'), ('flat', 'flat')], null=True),
        ),
    ]



import pytest


@pytest.mark.parallel
def test_prn_schema(pulp_openapi_schema):
    """Test that PRN is a part of every serializer with a pulp_href."""
    failed = []
    for name, schema in pulp_openapi_schema["components"]["schemas"].items():
        if name.endswith("Response"):
            if "pulp_href" in schema["properties"]:
                if "prn" in schema["properties"]:
                    prn_schema = schema["properties"]["prn"]
                    if prn_schema["type"] == "string" and prn_schema["readOnly"]:
                        continue
                failed.append(name)

    assert len(failed) == 0



from gettext import gettext as _
import sys

from django.conf import settings
from django.core.management import BaseCommand, CommandError

from pulp_rpm.app.models import Package  # noqa


class Command(BaseCommand):
    """
    Django management command for trimming changelog entries on packages in the Pulp database.

    RPM repositories contain a list of changelog entries for each package so that commands such
    as "dnf changelog" or "dnf repoquery --changelogs" can display it. Since this metadata can
    grow to be extremely large, typically you want to keep only a few changelogs for each package
    (if the package is actually installed on your system, you can view all changelogs with
    "rpm -qa $package --changelogs").

    In pulp_rpm 3.17 a new setting was introduced to enact this changelog limit, but it is not
    retroactively applied to packages that are already synced. This command will do so and can
    save a significant amount of disk space if Pulp is being used to sync RPM content from RHEL
    or Oracle Linux.
    """

    help = _(__doc__)

    def add_arguments(self, parser):
        """Set up arguments."""
        parser.add_argument(
            "--changelog-limit",
            default=settings.KEEP_CHANGELOG_LIMIT,
            type=int,
            required=False,
            help=_(
                "The number of changelogs you wish to keep for each package - the suggested "
                "value is 10. If this value is not provided the default configured as per the "
                "settings will be used."
            ),
        )

    def handle(self, *args, **options):
        """Implement the command."""
        changelog_limit = options["changelog_limit"]
        if changelog_limit <= 0:
            raise CommandError("--changelog-limit must be a non-zero positive integer")
        trimmed_packages = 0
        batch = []

        def update_total(total):
            sys.stdout.write("\rTrimmed changelogs for {} packages".format(total))
            sys.stdout.flush()

        for package in Package.objects.all().only("changelogs").iterator():
            # make sure the changelogs are ascending sorted by date
            package.changelogs.sort(key=lambda t: t[1])
            # take the last N changelogs
            package.changelogs = package.changelogs[-changelog_limit:]
            batch.append(package)
            if len(batch) > 500:
                Package.objects.bulk_update(batch, fields=["changelogs"])
                trimmed_packages += len(batch)
                batch.clear()
                update_total(trimmed_packages)

        Package.objects.bulk_update(batch, fields=["changelogs"])
        trimmed_packages += len(batch)
        batch.clear()
        update_total(trimmed_packages)
        print()



# Generated by Django 4.2.7 on 2023-12-12 18:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0058_alter_addon_repository_alter_variant_repository'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmpublication',
            name='compression_type',
            field=models.TextField(choices=[('zstd', 'zstd'), ('gz', 'gz')], null=True),
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='compression_type',
            field=models.TextField(choices=[('zstd', 'zstd'), ('gz', 'gz')], null=True),
        ),
    ]



"""Utilities for tests for the rpm plugin."""

import gzip
import os
import subprocess

import pyzstd
import requests

from pulp_rpm.tests.functional.constants import (
    PRIVATE_GPG_KEY_URL,
    PACKAGES_DIRECTORY,
)


def gen_rpm_content_attrs(artifact, rpm_name):
    """Generate a dict with content unit attributes.

    :param artifact: A dict of info about the artifact.
    :returns: A semi-random dict for use in creating a content unit.
    """
    return {"artifact": artifact.pulp_href, "relative_path": rpm_name}


def init_signed_repo_configuration():
    """Initialize the configuration required for verifying a signed repository.

    This function downloads and imports a private GPG key by invoking subprocess
    commands. Then, it creates a new signing service on the fly.
    """
    # download the private key
    priv_key = subprocess.run(
        ("wget", "-q", "-O", "-", PRIVATE_GPG_KEY_URL), stdout=subprocess.PIPE
    ).stdout
    # import the downloaded private key
    subprocess.run(("gpg", "--import"), input=priv_key)

    # set the imported key to the maximum trust level
    key_fingerprint = "0C1A894EBB86AFAE218424CADDEF3019C2D4A8CF"
    completed_process = subprocess.run(("echo", f"{key_fingerprint}:6:"), stdout=subprocess.PIPE)
    subprocess.run(("gpg", "--import-ownertrust"), input=completed_process.stdout)

    # create a new signing service
    utils_dir_path = os.path.dirname(os.path.realpath(__file__))
    signing_script_path = os.path.join(utils_dir_path, "sign-metadata.sh")

    return subprocess.run(
        (
            "pulpcore-manager",
            "add-signing-service",
            "sign-metadata",
            f"{signing_script_path}",
            "pulp-fixture-signing-key",
        )
    )


def get_package_repo_path(package_filename):
    """Get package repo path with directory structure.

    Args:
        package_filename(str): filename of RPM package

    Returns:
        (str): full path of RPM package in published repository

    """
    return os.path.join(PACKAGES_DIRECTORY, package_filename.lower()[0], package_filename)


def download_and_decompress_file(url):
    # Tests work normally but fails for S3 due '.gz'
    # Why is it only compressed for S3?
    resp = requests.get(url)
    decompression = None
    if url.endswith(".gz"):
        decompression = gzip.decompress
    elif url.endswith(".zst"):
        decompression = pyzstd.decompress

    if decompression:
        return decompression(resp.content)
    else:
        # FIXME: fix this as in CI primary/update_info.xml has '.gz' but it is not gzipped
        return resp.content



from pulpcore.plugin.viewsets import (
    ContentFilter,
    NoArtifactContentUploadViewSet,
)

from pulp_rpm.app.models import (
    UpdateRecord,
)
from pulp_rpm.app.serializers import (
    MinimalUpdateRecordSerializer,
    UpdateRecordSerializer,
)


class UpdateRecordFilter(ContentFilter):
    """
    FilterSet for UpdateRecord.
    """

    class Meta:
        model = UpdateRecord
        fields = {
            "id": ["exact", "in"],
            "status": ["exact", "in", "ne"],
            "severity": ["exact", "in", "ne"],
            "type": ["exact", "in", "ne"],
        }


class UpdateRecordViewSet(NoArtifactContentUploadViewSet):
    """
    A ViewSet for UpdateRecord.

    Define endpoint name which will appear in the API endpoint for this content type.
    For example::
        http://pulp.example.com/pulp/api/v3/content/rpm/advisories/

    Also specify queryset and serializer for UpdateRecord.
    """

    endpoint_name = "advisories"
    queryset = UpdateRecord.objects.all()
    serializer_class = UpdateRecordSerializer
    minimal_serializer_class = MinimalUpdateRecordSerializer
    filterset_class = UpdateRecordFilter

    # TODO: adjust this policy after upload access policy design done and in place
    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_required_repo_perms_on_upload:rpm.modify_content_rpmrepository",
                    "has_required_repo_perms_on_upload:rpm.view_rpmrepository",
                ],
            },
            {
                "action": ["set_label", "unset_label"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_domain_perms:core.manage_content_labels",
                ],
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }



from gettext import gettext as _

from rest_framework import fields, serializers

from pulp_rpm.app.models import RpmRepository

from pulpcore.plugin.serializers import ValidateFieldsMixin
from pulpcore.plugin.util import get_domain


class PrunePackagesSerializer(serializers.Serializer, ValidateFieldsMixin):
    """
    Serializer for prune-old-Packages operation.
    """

    repo_hrefs = fields.ListField(
        required=True,
        help_text=_(
            "Will prune old packages from the specified list of repos. "
            "Use ['*'] to specify all repos. "
            "Will prune based on the specified repositories' latest_versions."
        ),
        child=serializers.CharField(),
    )

    keep_days = serializers.IntegerField(
        help_text=_(
            "Prune packages introduced *prior-to* this many days ago. "
            "Default is 14. A value of 0 implies 'keep latest package only.'"
        ),
        required=False,
        min_value=0,
        default=14,
    )

    dry_run = serializers.BooleanField(
        help_text=_(
            "Determine what would-be-pruned and log the list of packages. "
            "Intended as a debugging aid."
        ),
        default=False,
        required=False,
    )

    def validate_repo_hrefs(self, value):
        """
        Insure repo_hrefs is not empty and contains either valid RPM Repository hrefs or "*".
        Args:
            value (list): The list supplied by the user
        Returns:
            The list of RpmRepositories after validation
        Raises:
            ValidationError: If the list is empty or contains invalid hrefs.
        """
        if len(value) == 0:
            raise serializers.ValidationError("Must not be [].")

        # prune-all-repos is "*" - find all RPM repos in this domain
        if "*" in value:
            if len(value) != 1:
                raise serializers.ValidationError("Can't specify specific HREFs when using '*'")
            return RpmRepository.objects.filter(pulp_domain=get_domain())

        from pulpcore.plugin.viewsets import NamedModelViewSet

        # We're pruning a specific list of RPM repositories.
        # Validate that they are for RpmRepositories.
        hrefs_to_return = []
        for href in value:
            hrefs_to_return.append(NamedModelViewSet.get_resource(href, RpmRepository))

        return hrefs_to_return



import pytest
import uuid

from pulpcore.client.pulp_rpm import RpmRepositorySyncURL
from pulpcore.client.pulp_rpm.exceptions import ApiException

from pulp_rpm.tests.functional.constants import (
    RPM_SIGNED_FIXTURE_URL,
    RPM_UNSIGNED_FIXTURE_URL,
    RPM_FIXTURE_SUMMARY,
)


@pytest.mark.parallel
def test_rbac_repositories(gen_user, rpm_repository_factory, rpm_repository_api, monitor_task):
    """
    Test creation of repository with user with permissions and without.
    """
    user_creator = gen_user(model_roles=["rpm.rpmrepository_creator", "rpm.rpmremote_creator"])
    user_viewer = gen_user(model_roles=["rpm.viewer"])
    user_no = gen_user()

    repo = None

    # Create repository
    with user_creator:
        repo = rpm_repository_factory()
        assert rpm_repository_api.read(repo.pulp_href)

    with user_viewer, pytest.raises(ApiException):
        rpm_repository_factory()

    with user_no, pytest.raises(ApiException):
        rpm_repository_factory()

    # List & retrieve repository
    with user_creator:
        assert rpm_repository_api.list(name=repo.name).count == 1
        assert repo.pulp_href == rpm_repository_api.read(repo.pulp_href).pulp_href

    with user_viewer:
        assert rpm_repository_api.list(name=repo.name).count == 1
        assert repo.pulp_href == rpm_repository_api.read(repo.pulp_href).pulp_href

    with user_no, pytest.raises(ApiException) as exc:
        assert rpm_repository_api.list(name=repo.name).count == 0
        rpm_repository_api.read(repo.pulp_href)
    assert exc.value.status == 404

    # Update repository
    with user_creator:
        repo_data = repo.to_dict()
        repo_data.update(name="rpm_repo_test_modify")
        monitor_task(rpm_repository_api.update(repo.pulp_href, repo_data).task)
        assert rpm_repository_api.read(repo.pulp_href).name == "rpm_repo_test_modify"

    with user_no, pytest.raises(ApiException) as exc:
        repo_data = repo.to_dict()
        repo_data.update(name="rpm_repo_test_modify_without_perms")
        rpm_repository_api.update(repo.pulp_href, repo_data)
    # Here is response `404` as user doesn't have even permission to retrieve repo data
    # so pulp response with not found instead `access denied` to not expose it exists
    assert exc.value.status == 404

    with user_viewer, pytest.raises(ApiException) as exc:
        repo_data = repo.to_dict()
        repo_data.update(name="rpm_repo_test_modify_with_view_perms")
        rpm_repository_api.update(repo.pulp_href, repo_data)
    # Fails with '403' as a repo can be seen but not updated.
    assert exc.value.status == 403

    # Remove repository
    # Fails with `404` not found for same reason as `update` above
    with user_no, pytest.raises(ApiException) as exc:
        rpm_repository_api.delete(repo.pulp_href)
    assert exc.value.status == 404

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_repository_api.delete(repo.pulp_href)
    assert exc.value.status == 403

    with user_creator:
        response = rpm_repository_api.delete(repo.pulp_href)
        monitor_task(response.task)
        assert rpm_repository_api.list(name=repo.name).count == 0


@pytest.mark.parallel
def test_rbac_remotes_and_sync(
    gen_user,
    rpm_rpmremote_api,
    rpm_repository_api,
    rpm_repository_factory,
    rpm_rpmremote_factory,
    monitor_task,
):
    """
    Test creation of remotes with user with permissions and without.

    Then to test sync of one of those remote.
    """
    user_creator = gen_user(
        model_roles=[
            "rpm.rpmremote_creator",
            "rpm.rpmrepository_creator",
        ]
    )
    user_viewer = gen_user(model_roles=["rpm.viewer"])
    user_no = gen_user(model_roles=["rpm.rpmrepository_creator"])

    remote = None

    # Create
    with user_no, pytest.raises(ApiException) as exc:
        rpm_rpmremote_factory(url=RPM_SIGNED_FIXTURE_URL)
    assert exc.value.status == 403

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_rpmremote_factory(url=RPM_SIGNED_FIXTURE_URL)
    assert exc.value.status == 403

    with user_creator:
        remote = rpm_rpmremote_factory(url=RPM_SIGNED_FIXTURE_URL)
        assert rpm_rpmremote_api.list(name=remote.name).count == 1

    # Update
    remote_data_update = remote.to_dict()
    remote_data_update["url"] = RPM_UNSIGNED_FIXTURE_URL

    with user_no, pytest.raises(ApiException) as exc:
        rpm_rpmremote_api.update(remote.pulp_href, remote_data_update)
    assert exc.value.status == 404

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_rpmremote_api.update(remote.pulp_href, remote_data_update)
    # 403 as user can view content but not update
    assert exc.value.status == 403

    with user_creator:
        res = rpm_rpmremote_api.update(remote.pulp_href, remote_data_update)
        monitor_task(res.task)
        remote = rpm_rpmremote_api.read(remote.pulp_href)
        remotes = rpm_rpmremote_api.list(name=remote_data_update["name"]).results
        assert len(remotes) == 1
        assert remotes[0].url == RPM_UNSIGNED_FIXTURE_URL

    # Sync the remote
    with user_no, pytest.raises(ApiException) as exc:
        repo = rpm_repository_factory(remote=remote.pulp_href)
        sync_url = RpmRepositorySyncURL(remote=remote.pulp_href)
        rpm_repository_api.sync(repo.pulp_href, sync_url)
    assert exc.value.status == 403

    with user_viewer, pytest.raises(ApiException) as exc:
        repo = rpm_repository_factory(remote=remote.pulp_href)
        sync_url = RpmRepositorySyncURL(remote=remote.pulp_href)
        rpm_repository_api.sync(repo.pulp_href, sync_url)
    assert exc.value.status == 403

    with user_creator:
        repo = rpm_repository_factory(remote=remote.pulp_href)
        sync_url = RpmRepositorySyncURL(remote=remote.pulp_href)
        sync_res = rpm_repository_api.sync(repo.pulp_href, sync_url)
        monitor_task(sync_res.task)
        repo = rpm_repository_api.read(repo.pulp_href)
        assert repo.latest_version_href.endswith("/1/")

    # Remove
    with user_no, pytest.raises(ApiException) as exc:
        rpm_repository_api.delete(remote.pulp_href)
    assert exc.value.status == 404

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_repository_api.delete(remote.pulp_href)
    assert exc.value.status == 403

    with user_creator:
        response = rpm_rpmremote_api.delete(remote.pulp_href)
        monitor_task(response.task)
        response = rpm_repository_api.delete(repo.pulp_href)
        monitor_task(response.task)
        assert rpm_rpmremote_api.list(name=remote.name).count == 0
        assert rpm_repository_api.list(name=repo.name).count == 0


@pytest.mark.parallel
def test_rbac_acs(gen_user, rpm_acs_api, rpm_rpmremote_api, rpm_rpmremote_factory, monitor_task):
    """Test RPM ACS CRUD."""
    user_creator = gen_user(
        model_roles=[
            "rpm.rpmalternatecontentsource_creator",
            "rpm.rpmremote_owner",
        ]
    )
    user_viewer = gen_user(
        model_roles=[
            "rpm.viewer",
            "rpm.rpmremote_owner",
        ]
    )
    user_no = gen_user(
        model_roles=[
            "rpm.rpmremote_owner",
        ]
    )

    acs = None
    remote = rpm_rpmremote_factory(policy="on_demand")

    acs_data = {
        "name": str(uuid.uuid4()),
        "remote": remote.pulp_href,
    }

    # Create
    with user_no, pytest.raises(ApiException) as exc:
        rpm_acs_api.create(acs_data)
    assert exc.value.status == 403

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_acs_api.create(acs_data)
    assert exc.value.status == 403

    with user_creator:
        acs = rpm_acs_api.create(acs_data)
        assert rpm_acs_api.list(name=acs.name).count == 1

    # Update & Read
    with user_no, pytest.raises(ApiException) as exc:
        rpm_acs_api.read(acs.pulp_href)
    assert exc.value.status == 404

    with user_viewer, pytest.raises(ApiException) as exc:
        acs_to_update = rpm_acs_api.read(acs.pulp_href).to_dict()
        acs_to_update["paths"][0] = "files/"
        rpm_acs_api.update(acs.pulp_href, acs_to_update)
    assert exc.value.status == 403

    with user_creator:
        acs_to_update = rpm_acs_api.read(acs.pulp_href).to_dict()
        acs_to_update["paths"][0] = "files/"
        response = rpm_acs_api.update(acs.pulp_href, acs_to_update)
        monitor_task(response.task)
        assert rpm_acs_api.list(name=acs.name).count == 1
        assert "files/" in rpm_acs_api.read(acs.pulp_href).paths

    # Remove
    with user_no, pytest.raises(ApiException) as exc:
        rpm_acs_api.delete(acs.pulp_href)
    assert exc.value.status == 404

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_acs_api.delete(acs.pulp_href)
    assert exc.value.status == 403

    with user_creator:
        response = rpm_acs_api.delete(acs.pulp_href)
        monitor_task(response.task)
        response = rpm_rpmremote_api.delete(remote.pulp_href)
        monitor_task(response.task)
        assert rpm_acs_api.list(name=acs.name).count == 0


@pytest.mark.parallel
def test_rbac_publication(
    gen_user,
    rpm_rpmremote_api,
    rpm_repository_api,
    rpm_repository_factory,
    rpm_publication_factory,
    rpm_publication_api,
    monitor_task,
):
    """Test RPM publication CRUD."""
    user_creator = gen_user(
        model_roles=[
            "rpm.rpmpublication_creator",
            "rpm.rpmremote_creator",
            "rpm.rpmrepository_creator",
        ]
    )
    user_viewer = gen_user(
        model_roles=[
            "rpm.viewer",
        ]
    )
    user_no = gen_user(model_roles=[])

    # Create
    with user_creator:
        repo = rpm_repository_factory()
        publication = rpm_publication_factory(repository=repo.pulp_href)
        assert publication.repository == repo.pulp_href
        pub_from_repo_version = rpm_publication_factory(repository_version=repo.latest_version_href)
        assert pub_from_repo_version.repository_version == repo.latest_version_href
    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_publication_factory(repository=repo.pulp_href)
    assert exc.value.status == 403

    with user_no, pytest.raises(ApiException) as exc:
        rpm_publication_factory(repository=repo.pulp_href)
    assert exc.value.status == 403

    # Remove
    with user_no, pytest.raises(ApiException) as exc:
        rpm_publication_api.delete(publication.pulp_href)
    assert exc.value.status == 404

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_publication_api.delete(publication.pulp_href)
    assert exc.value.status == 403

    with user_creator:
        rpm_publication_api.delete(publication.pulp_href)
        res = rpm_repository_api.delete(repo.pulp_href)
        monitor_task(res.task)
        publications = rpm_publication_api.list().results
        assert not any(p.repository != repo.pulp_href for p in publications)


@pytest.mark.parallel
def test_rbac_distribution(
    gen_user,
    rpm_repository_api,
    rpm_repository_factory,
    rpm_rpmremote_api,
    rpm_publication_api,
    rpm_publication_factory,
    rpm_distribution_api,
    rpm_distribution_factory,
    monitor_task,
):
    """Test RPM distribution CRUD."""
    user_creator = gen_user(
        model_roles=[
            "rpm.rpmrepository_creator",
            "rpm.rpmpublication_creator",
            "rpm.rpmdistribution_creator",
        ]
    )
    user_viewer = gen_user(
        model_roles=[
            "rpm.viewer",
        ]
    )
    user_no = gen_user(model_roles=[])

    # Create
    with user_no, pytest.raises(ApiException) as exc:
        rpm_distribution_factory(name=str(uuid.uuid4()), base_path=str(uuid.uuid4()))
    assert exc.value.status == 403

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_distribution_factory(name=str(uuid.uuid4()), base_path=str(uuid.uuid4()))
    assert exc.value.status == 403

    with user_creator:
        repo = rpm_repository_factory()
        publication = rpm_publication_factory(repository=repo.pulp_href)
        distribution = rpm_distribution_factory(publication=publication.pulp_href)
        assert distribution.publication == publication.pulp_href

    # Update
    dist_data_to_update = rpm_distribution_api.read(distribution.pulp_href).to_dict()
    new_name = str(uuid.uuid4())
    dist_data_to_update["name"] = new_name

    with user_no, pytest.raises(ApiException) as exc:
        rpm_distribution_api.update(distribution.pulp_href, dist_data_to_update)
    assert exc.value.status == 404

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_distribution_api.update(distribution.pulp_href, dist_data_to_update)
    assert exc.value.status == 403

    with user_creator:
        res = rpm_distribution_api.update(distribution.pulp_href, dist_data_to_update)
        monitor_task(res.task)
        assert rpm_distribution_api.list(name=new_name).count == 1

    # Remove
    with user_no, pytest.raises(ApiException) as exc:
        rpm_distribution_api.delete(distribution.pulp_href)
    assert exc.value.status == 404

    with user_viewer, pytest.raises(ApiException) as exc:
        rpm_distribution_api.delete(distribution.pulp_href)
    assert exc.value.status == 403

    with user_creator:
        rpm_distribution_api.delete(distribution.pulp_href)
        rpm_publication_api.delete(publication.pulp_href)

        res = rpm_repository_api.delete(repo.pulp_href)
        monitor_task(res.task)

        assert rpm_distribution_api.list(name=distribution.name).count == 0
        assert rpm_repository_api.list(name=repo.name).count == 0


@pytest.mark.parallel
def test_rbac_content_scoping(
    gen_user,
    rpm_advisory_api,
    rpm_package_api,
    rpm_package_category_api,
    rpm_package_groups_api,
    rpm_package_lang_packs_api,
    rpm_repository_api,
    rpm_repository_factory,
    rpm_rpmremote_api,
    rpm_rpmremote_factory,
    monitor_task,
):
    """
    Test creation of remotes with user with permissions and without.

    Then to test sync of one of those remote.
    """
    user_creator = gen_user(
        model_roles=[
            "rpm.rpmremote_creator",
            "rpm.rpmrepository_creator",
        ]
    )
    user_viewer = gen_user(model_roles=["rpm.viewer"])
    user_no = gen_user(model_roles=["rpm.rpmrepository_creator"])

    remote = None

    # Create
    with user_creator:
        remote = rpm_rpmremote_factory(url=RPM_SIGNED_FIXTURE_URL)
        assert rpm_rpmremote_api.list().count == 1

    # Sync the remote
    with user_creator:
        repo = rpm_repository_factory(remote=remote.pulp_href)
        sync_url = RpmRepositorySyncURL(remote=remote.pulp_href)
        sync_res = rpm_repository_api.sync(repo.pulp_href, sync_url)
        monitor_task(sync_res.task)
        repo = rpm_repository_api.read(repo.pulp_href)
        assert repo.latest_version_href.endswith("/1/")

    def _assert_listed_content():
        packages_count = rpm_package_api.list(repository_version=repo.latest_version_href).count
        assert RPM_FIXTURE_SUMMARY["rpm.package"]["count"] == packages_count

        advisories_count = rpm_advisory_api.list(repository_version=repo.latest_version_href).count
        assert RPM_FIXTURE_SUMMARY["rpm.advisory"]["count"] == advisories_count

        package_categories_count = rpm_package_category_api.list(
            repository_version=repo.latest_version_href
        ).count
        assert RPM_FIXTURE_SUMMARY["rpm.packagecategory"]["count"] == package_categories_count

        package_groups_count = rpm_package_groups_api.list(
            repository_version=repo.latest_version_href
        ).count
        assert RPM_FIXTURE_SUMMARY["rpm.packagegroup"]["count"] == package_groups_count

        package_lang_packs_count = rpm_package_lang_packs_api.list(
            repository_version=repo.latest_version_href
        ).count
        assert RPM_FIXTURE_SUMMARY["rpm.packagelangpacks"]["count"] == package_lang_packs_count

    # Test content visibility
    # TODO: modules
    with user_creator:
        _assert_listed_content()

    with user_viewer:
        _assert_listed_content()

    with user_no:
        assert 0 == rpm_package_api.list(repository_version=repo.latest_version_href).count
        assert 0 == rpm_advisory_api.list(repository_version=repo.latest_version_href).count
        assert 0 == rpm_package_category_api.list(repository_version=repo.latest_version_href).count
        assert 0 == rpm_package_groups_api.list(repository_version=repo.latest_version_href).count
        assert (
            0 == rpm_package_lang_packs_api.list(repository_version=repo.latest_version_href).count
        )

    # Remove
    with user_creator:
        response = rpm_rpmremote_api.delete(remote.pulp_href)
        monitor_task(response.task)
        response = rpm_repository_api.delete(repo.pulp_href)
        monitor_task(response.task)
        assert rpm_rpmremote_api.list(name=remote.name).count == 0
        assert rpm_repository_api.list(name=repo.name).count == 0



from django.db import transaction
from django_filters import CharFilter
from drf_spectacular.utils import extend_schema
from pulpcore.plugin.models import PulpTemporaryFile
from pulpcore.plugin.serializers import AsyncOperationResponseSerializer
from pulpcore.plugin.tasking import dispatch
from pulpcore.plugin.viewsets import (
    ContentFilter,
    OperationPostponedResponse,
    SingleArtifactContentUploadViewSet,
)
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response

from pulp_rpm.app import tasks as rpm_tasks
from pulp_rpm.app.models import Package
from pulp_rpm.app.serializers import (
    MinimalPackageSerializer,
    PackageSerializer,
    PackageUploadSerializer,
)


class PackageFilter(ContentFilter):
    """
    FilterSet for Package.
    """

    sha256 = CharFilter(field_name="_artifacts__sha256")
    filename = CharFilter(field_name="content_artifact__relative_path")

    class Meta:
        model = Package
        fields = {
            "name": ["exact", "in", "ne", "contains", "startswith"],
            "epoch": ["exact", "in", "ne"],
            "version": ["exact", "in", "ne"],
            "release": ["exact", "in", "ne", "contains", "startswith"],
            "arch": ["exact", "in", "ne", "contains", "startswith"],
            "pkgId": ["exact", "in"],
            "checksum_type": ["exact", "in", "ne"],
        }


class PackageViewSet(SingleArtifactContentUploadViewSet):
    """
    A ViewSet for Package.

    Define endpoint name which will appear in the API endpoint for this content type.
    For example::
        http://pulp.example.com/pulp/api/v3/content/rpm/packages/

    Also specify queryset and serializer for Package.
    """

    endpoint_name = "packages"
    queryset = Package.objects.prefetch_related("_artifacts")
    serializer_class = PackageSerializer
    minimal_serializer_class = MinimalPackageSerializer
    filterset_class = PackageFilter

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_required_repo_perms_on_upload:rpm.modify_content_rpmrepository",
                    "has_required_repo_perms_on_upload:rpm.view_rpmrepository",
                ],
            },
            {
                "action": ["upload"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_domain_perms:rpm.upload_rpm_packages",
                ],
            },
            {
                "action": ["set_label", "unset_label"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_domain_perms:core.manage_content_labels",
                ],
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }

    LOCKED_ROLES = {
        "rpm.rpm_package_uploader": [
            "rpm.upload_rpm_packages",
        ],
    }

    @extend_schema(
        description="Trigger an asynchronous task to create an RPM package,"
        "optionally create new repository version.",
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request):
        # validation decides if we want to sign and set that in the context space
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        if serializer.context["sign_package"] is False:
            return super().create(request)

        # signing case
        validated_data = serializer.validated_data
        signing_service_pk = validated_data["repository"].package_signing_service.pk
        signing_fingerprint = validated_data["repository"].package_signing_fingerprint
        if "file" in validated_data:
            request.data.pop("file")
            temp_uploaded_file = validated_data["file"]
            pulp_temp_file = PulpTemporaryFile(file=temp_uploaded_file.temporary_file_path())
            pulp_temp_file.save()
        else:
            pulp_temp_file = validated_data["upload"]

        # dispatch signing task
        task_args = {
            "app_label": self.queryset.model._meta.app_label,
            "serializer_name": serializer.__class__.__name__,
            "signing_service_pk": signing_service_pk,
            "signing_fingerprint": signing_fingerprint,
            "temporary_file_pk": pulp_temp_file.pk,
        }
        task_payload = {k: v for k, v in request.data.items()}
        task_exclusive = [
            serializer.validated_data.get("upload"),
            serializer.validated_data.get("repository"),
        ]
        task = dispatch(
            rpm_tasks.signing.sign_and_create,
            exclusive_resources=task_exclusive,
            args=tuple(task_args.values()),
            kwargs={
                "data": task_payload,
                "context": self.get_deferred_context(request),
            },
        )
        return OperationPostponedResponse(task, request)

    @extend_schema(
        description="Synchronously upload an RPM package.",
        request=PackageUploadSerializer,
        responses={201: PackageSerializer},
        summary="Upload an RPM package synchronously.",
    )
    @action(detail=False, methods=["post"], serializer_class=PackageUploadSerializer)
    def upload(self, request):
        """Create an RPM package."""
        serializer = self.get_serializer(data=request.data)
        with transaction.atomic():
            # Create the artifact
            serializer.is_valid(raise_exception=True)
            # Create the Package
            serializer.save()

        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)



from drf_spectacular.utils import extend_schema
from django.conf import settings
from rest_framework.viewsets import ViewSet

from pulpcore.plugin.viewsets import TaskGroupOperationResponse
from pulpcore.plugin.models import TaskGroup
from pulpcore.plugin.serializers import TaskGroupOperationResponseSerializer
from pulp_rpm.app.serializers import PrunePackagesSerializer
from pulp_rpm.app.tasks import prune_packages
from pulpcore.plugin.tasking import dispatch


class PrunePackagesViewSet(ViewSet):
    """
    Viewset for prune-old-Packages endpoint.
    """

    serializer_class = PrunePackagesSerializer

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["prune_packages"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_repository_model_or_domain_or_obj_perms:rpm.modify_content_rpmrepository",
                    "has_repository_model_or_domain_or_obj_perms:rpm.view_rpmrepository",
                ],
            },
        ],
    }

    @extend_schema(
        description="Trigger an asynchronous old-Package-prune operation.",
        responses={202: TaskGroupOperationResponseSerializer},
    )
    def prune_packages(self, request):
        """
        Triggers an asynchronous old-Package-purge operation.

        This returns a task-group that contains a "master" task that dispatches one task
        per repo being pruned. This allows repositories to become available for other
        processing as soon as their task completes, rather than having to wait for *all*
        repositories to be pruned.
        """
        serializer = PrunePackagesSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        repos = serializer.validated_data.get("repo_hrefs", [])
        repos_to_prune_pks = []
        for repo in repos:
            repos_to_prune_pks.append(repo.pk)

        uri = "/api/v3/rpm/prune/"
        if settings.DOMAIN_ENABLED:
            uri = f"/{request.pulp_domain.name}{uri}"
        exclusive_resources = [uri, f"pdrn:{request.pulp_domain.pulp_id}:rpm:prune"]

        task_group = TaskGroup.objects.create(description="Prune old Packages.")

        dispatch(
            prune_packages,
            exclusive_resources=exclusive_resources,
            task_group=task_group,
            kwargs={
                "repo_pks": repos_to_prune_pks,
                "keep_days": serializer.validated_data["keep_days"],
                "dry_run": serializer.validated_data["dry_run"],
            },
        )
        return TaskGroupOperationResponse(task_group, request)



"""Tests for Pulp`s download policies."""

import pytest

from pulp_rpm.tests.functional.constants import (
    RPM_FIXTURE_SUMMARY,
    DOWNLOAD_POLICIES,
)
from pulpcore.client.pulp_rpm import RpmRpmPublication


@pytest.mark.parametrize("download_policy", DOWNLOAD_POLICIES)
def test_download_policies(
    download_policy,
    init_and_sync,
    rpm_repository_version_api,
    rpm_publication_api,
    gen_object_with_cleanup,
    delete_orphans_pre,
    get_content_summary,
):
    """Sync repositories with the different ``download_policy``.

    Do the following:

    1. Create a repository, and a remote.
    2. Sync the remote.
    3. Assert that repository version is not None.
    4. Assert that the correct number of possible units to be downloaded
       were shown.
    5. Sync again with the same remote.
    6. Assert that the latest repository version did not change.
    7. Assert that the same number of units are shown, and after the
       second sync no extra units should be shown, since the same remote
       was synced again.
    8. Publish repository synced with lazy ``download_policy``.
    """
    # Step 1, 2
    repo, remote = init_and_sync(policy=download_policy)

    # Step 3, 4
    assert repo.latest_version_href.endswith("/1/")
    content_summary = get_content_summary(repo)
    assert content_summary["present"] == RPM_FIXTURE_SUMMARY
    assert content_summary["added"] == RPM_FIXTURE_SUMMARY

    # Step 5
    latest_version_href = repo.latest_version_href
    repo, remote = init_and_sync(repository=repo, remote=remote)

    # Step 6, 7
    assert latest_version_href == repo.latest_version_href
    content_summary = get_content_summary(repo)
    assert content_summary["present"] == RPM_FIXTURE_SUMMARY

    # Step 8
    publish_data = RpmRpmPublication(repository=repo.pulp_href)
    publication = gen_object_with_cleanup(rpm_publication_api, publish_data)

    assert publication.repository is not None
    assert publication.repository_version is not None



import pytest

from pulpcore.client.pulp_rpm import PrunePackages
from pulpcore.client.pulp_rpm.exceptions import ApiException


def test_01_prune_params(init_and_sync, rpm_prune_api, monitor_task_group):
    """Assert on various param-validation errors."""
    # create/sync rpm repo
    repo, _ = init_and_sync(policy="on_demand")

    params = PrunePackages(repo_hrefs=[])
    # requires repo-href or *
    with pytest.raises(ApiException) as exc:
        rpm_prune_api.prune_packages(params)
    assert "Must not be []" in exc.value.body

    params.repo_hrefs = ["foo"]
    with pytest.raises(ApiException) as exc:
        rpm_prune_api.prune_packages(params)
    assert "URI not valid" in exc.value.body

    params.repo_hrefs = ["*", repo.pulp_href]
    with pytest.raises(ApiException) as exc:
        rpm_prune_api.prune_packages(params)
    assert "Can't specify specific HREFs when using" in exc.value.body

    # '*' only'
    params.repo_hrefs = ["*"]
    params.dry_run = True
    params.keep_days = 1000
    task_group = monitor_task_group(rpm_prune_api.prune_packages(params).task_group)
    assert 0 == task_group.failed

    # Valid repo-href
    params.repo_hrefs = [repo.pulp_href]
    task_group = monitor_task_group(rpm_prune_api.prune_packages(params).task_group)
    assert 2 == task_group.completed
    assert 0 == task_group.failed

    # Valid +int
    params.keep_days = 1000
    task_group = monitor_task_group(rpm_prune_api.prune_packages(params).task_group)
    assert 2 == task_group.completed
    assert 0 == task_group.failed

    # Valid 0
    params.keep_days = 0
    task_group = monitor_task_group(rpm_prune_api.prune_packages(params).task_group)
    assert 2 == task_group.completed
    assert 0 == task_group.failed

    # Valid dry-run
    params.dry_run = True
    task_group = monitor_task_group(rpm_prune_api.prune_packages(params).task_group)
    assert 2 == task_group.completed
    assert 0 == task_group.failed


def test_02_prune_dry_run(init_and_sync, rpm_prune_api, monitor_task_group, monitor_task):
    # create/sync rpm repo
    repo, _ = init_and_sync(policy="on_demand")

    # prune keep=0 dry_run=True -> expect total=4 done=0
    params = PrunePackages(repo_hrefs=[repo.pulp_href], keep_days=0, dry_run=True)
    task_group = monitor_task_group(rpm_prune_api.prune_packages(params).task_group)
    assert 2 == len(task_group.tasks)
    assert 2 == task_group.completed
    assert 0 == task_group.failed
    assert 1 == len(task_group.group_progress_reports)

    prog_rpt = task_group.group_progress_reports[0]
    assert 1 == prog_rpt.done
    assert 1 == prog_rpt.total
    for t in task_group.tasks:
        if t.name == "pulp_rpm.app.tasks.prune.prune_repo_packages":
            prune_task = monitor_task(t.pulp_href)
            assert 1 == len(prune_task.progress_reports)
            assert 4 == prune_task.progress_reports[0].total
            assert 0 == prune_task.progress_reports[0].done

    # prune keep=1000 dry_run=True -> expect total=0 done=0
    params = PrunePackages(repo_hrefs=[repo.pulp_href], keep_days=1000, dry_run=True)
    task_group = monitor_task_group(rpm_prune_api.prune_packages(params).task_group)
    assert 2 == len(task_group.tasks)
    assert 2 == task_group.completed
    assert 0 == task_group.failed
    for t in task_group.tasks:
        if t.name == "pulp_rpm.app.tasks.prune.prune_repo_packages":
            prune_task = monitor_task(t.pulp_href)
            assert 1 == len(prune_task.progress_reports)
            assert 0 == prune_task.progress_reports[0].total
            assert 0 == prune_task.progress_reports[0].done


def test_03_prune_results(
    init_and_sync,
    rpm_prune_api,
    monitor_task_group,
    monitor_task,
    rpm_repository_api,
    rpm_repository_version_api,
):
    # create/sync rpm repo
    repo, _ = init_and_sync(policy="on_demand")

    # prune keep=0 dry_run=False -> expect total=4 done=4
    params = PrunePackages(repo_hrefs=[repo.pulp_href], keep_days=0, dry_run=False)
    task_group = monitor_task_group(rpm_prune_api.prune_packages(params).task_group)
    assert 2 == len(task_group.tasks)
    assert 2 == task_group.completed
    assert 0 == task_group.failed

    for t in task_group.tasks:
        if t.name == "pulp_rpm.app.tasks.prune.prune_repo_packages":
            prune_task = monitor_task(t.pulp_href)
            assert 1 == len(prune_task.progress_reports)
            assert 4 == prune_task.progress_reports[0].total
            assert 4 == prune_task.progress_reports[0].done

    # investigate content -> 4 fewer packages, correct dups gone
    repo2 = rpm_repository_api.read(repo.pulp_href)
    rv = rpm_repository_version_api.read(repo2.latest_version_href)
    assert 4 == rv.content_summary.removed["rpm.package"]["count"]



from pulp_rpm.app.modulemd import parse_modular
import os

sample_file_data = """
---
document: modulemd
version: 2
data:
  name: kangaroo
  stream: 1.10
  version: 20180730223407
  context: deadbeef
  static_context: true
  arch: noarch
  summary: Kangaroo 0.3 module
  description: >-
    A module for the kangaroo 0.3 package
  license:
    module:
      - MIT
    content:
      - MIT
  profiles:
    default:
      rpms:
        - kangaroo
  artifacts:
    rpms:
      - kangaroo-0:0.3-1.noarch
...
---
document: modulemd
version: 2
data:
  name: kangaroo
  stream: "1.10"
  version: 20180704111719
  context: deadbeef
  static_context: false
  arch: noarch
  summary: Kangaroo 0.2 module
  description: >-
    A module for the kangaroo 0.2 package
  license:
    module:
      - MIT
    content:
      - MIT
  profiles:
    default:
      rpms:
        - kangaroo
  artifacts:
    rpms:
      - kangaroo-0:0.2-1.noarch
...
---
document: modulemd-defaults
version: 1
data:
  module: avocado
  modified: 202002242100
  profiles:
    "5.30": [default]
    "stream": [default]
...
---
document: modulemd-obsoletes
version: 1
data:
  modified: 2022-01-24T08:54Z
  module: perl
  stream: 5.30
  eol_date: 2021-06-01T00:00Z
  message: Module stream perl:5.30 is no longer supported. Please switch to perl:5.32
  obsoleted_by:
    module: perl
    stream: 5.40
...
"""


def test_parse_modular_preserves_literal_unquoted_values_3285(tmp_path):
    """Unquoted yaml numbers with trailing/leading zeros are preserved on specific fields"""

    # write data to test_file
    os.chdir(tmp_path)
    file_name = "modulemd.yaml"
    with open(file_name, "w") as file:
        file.write(sample_file_data)

    all, defaults, obsoletes = parse_modular(file_name)

    # check normal, defaults and obsoletes modulemds
    kangoroo1 = all[0]  # unquoted
    kangoroo2 = all[1]  # quoted
    modulemd_defaults = defaults[0]
    modulemd_obsoletes = obsoletes[0]

    assert kangoroo1["name"] == "kangaroo"
    assert kangoroo1["stream"] == "1.10"  # should not be 1.1
    assert kangoroo1["version"] == "20180730223407"
    assert kangoroo1["context"] == "deadbeef"
    assert kangoroo1["static_context"] is True
    assert kangoroo1["arch"] == "noarch"

    assert kangoroo2["name"] == "kangaroo"
    assert kangoroo2["stream"] == "1.10"
    assert kangoroo2["version"] == "20180704111719"
    assert kangoroo2["context"] == "deadbeef"
    assert kangoroo2["static_context"] is False
    assert kangoroo2["arch"] == "noarch"

    # 'stream' keys which have non-scalar values (e.g. list) are parsed normally.
    # Otherwise, weird results are produced (internal pyyaml objects)
    assert modulemd_defaults["module"] == "avocado"
    assert modulemd_defaults.get("modified") is None  # not present
    assert modulemd_defaults["profiles"]["stream"] == ["default"]
    assert modulemd_defaults["profiles"]["5.30"] == ["default"]

    # parse_modular changes the structure and key names for obsoletes
    assert modulemd_obsoletes["modified"] == "2022-01-24T08:54Z"
    assert modulemd_obsoletes["module_name"] == "perl"
    assert modulemd_obsoletes["module_stream"] == "5.30"
    assert modulemd_obsoletes["eol_date"] == "2021-06-01T00:00Z"
    assert (
        modulemd_obsoletes["message"]
        == "Module stream perl:5.30 is no longer supported. Please switch to perl:5.32"
    )
    assert modulemd_obsoletes["obsoleted_by_module_name"] == "perl"
    assert modulemd_obsoletes["obsoleted_by_module_stream"] == "5.40"



import socket
from pathlib import Path

import pytest
import requests
from pulpcore.plugin.exceptions import InvalidSignatureError

from pulp_rpm.app.shared_utils import RpmTool
from pulp_rpm.tests.functional.constants import PUBLIC_GPG_KEY_URL, RPM_SIGNED_URL, RPM_UNSIGNED_URL


def connection_guard(*args, **kwargs):
    raise Exception("I told you not to use the Internet!")


def get_fixture(tmp_path: Path, url: str):
    """Utility to get unsigned package"""
    result = requests.get(url)
    file = tmp_path / result.url.split("/")[-1]
    file.write_bytes(result.content)
    return file


def test_can_get_empty_rpm(tmp_path, monkeypatch):
    """
    Can get a valid rpm without hitting the internet.

    This rpm can be used in production by the SigningService.validate() method, which
    is used to validate a provided signing script knows how to sign an rpm blob.
    """
    # Should't hit the internet
    # https://stackoverflow.com/a/18601897
    monkeypatch.setattr(socket, "socket", connection_guard)
    rpm_pkg = RpmTool.get_empty_rpm(tmp_path)
    assert rpm_pkg.exists()


def test_verify_signature_is_valid(tmp_path):
    """Can verify that a package is unsigned"""
    pkg_file = get_fixture(tmp_path, RPM_SIGNED_URL)
    pubkey = get_fixture(tmp_path, PUBLIC_GPG_KEY_URL)

    rpm_tool = RpmTool(tmp_path)
    rpm_tool.import_pubkey_file(pubkey)
    rpm_tool.verify_signature(pkg_file)


def test_verify_package_is_unsigned(tmp_path):
    """Can verify that a package is unsigned"""
    pkg_file = get_fixture(tmp_path, RPM_UNSIGNED_URL)

    rpm_tool = RpmTool(tmp_path)
    with pytest.raises(InvalidSignatureError, match=RpmTool.UNSIGNED_ERROR_MSG):
        rpm_tool.verify_signature(pkg_file)


def test_verify_signature_is_invalid(tmp_path):
    """Can verify that a package's signature is invalid.
    That means that no pubkeys in the rpm db can validate the package.
    """
    pkg_file = get_fixture(tmp_path, RPM_SIGNED_URL)

    rpm_tool = RpmTool(tmp_path)
    with pytest.raises(InvalidSignatureError, match=RpmTool.INVALID_SIGNATURE_ERROR_MSG):
        rpm_tool.verify_signature(pkg_file)


def test_alternative_root_works(tmp_path):
    """Can use alternative "rpm --root" option for isolated operations."""
    pkg_file = get_fixture(tmp_path, RPM_SIGNED_URL)
    pubkey = get_fixture(tmp_path, PUBLIC_GPG_KEY_URL)

    # 1. First instance imports pubkey using a custom root
    root_1 = tmp_path / "root_1"
    rpm_tool_1 = RpmTool(root=root_1)
    rpm_tool_1.import_pubkey_file(str(pubkey))
    rpm_tool_1.verify_signature(pkg_file)

    # 2. Second instance uses a different root:
    # Because it doesnt have access to the other db (in root_1),
    # it raises for not being able to find the pubkey to verify the pkg
    root_2 = tmp_path / "root_2"
    rpm_tool_2 = RpmTool(root=root_2)
    with pytest.raises(InvalidSignatureError, match=RpmTool.INVALID_SIGNATURE_ERROR_MSG):
        rpm_tool_2.verify_signature(pkg_file)



import shutil
import subprocess
import tempfile
import typing as t
from hashlib import sha256
from pathlib import Path

import createrepo_c as cr
from django.conf import settings
from django.utils.dateparse import parse_datetime
from importlib_resources import files
from pulpcore.plugin.exceptions import InvalidSignatureError


def format_nevra(name=None, epoch=0, version=None, release=None, arch=None):
    """Generate Name-Epoch-Version-Release-Arch string."""
    return "{name}-{epoch}:{version}-{release}.{arch}".format(
        name=name, epoch=epoch, version=version, release=release, arch=arch
    )


def format_nvra(name=None, version=None, release=None, arch=None):
    """Generate Name-Version-Release-Arch from a package metadata."""
    return "{name}-{version}-{release}.{arch}".format(
        name=name,
        version=version,
        release=release,
        arch=arch,
    )


def format_nevra_short(name=None, epoch=0, version=None, release=None, arch=None):
    """Returns NEVRA or NVRA based on epoch."""
    if int(epoch) > 0:
        return format_nevra(name, epoch, version, release, arch)
    return format_nvra(name, version, release, arch)


def read_crpackage_from_artifact(artifact):
    """
    Helper function for creating package.

    Copy file to a temp directory and parse it.

    Returns: package model as dict

    Args:
        artifact: inited and validated artifact to save
    """
    filename = f"{artifact.pulp_id}.rpm"
    artifact_file = artifact.pulp_domain.get_storage().open(artifact.file.name)
    with tempfile.NamedTemporaryFile("wb", dir=".", suffix=filename) as temp_file:
        shutil.copyfileobj(artifact_file, temp_file)
        temp_file.flush()
        cr_pkginfo = cr.package_from_rpm(
            temp_file.name, changelog_limit=settings.KEEP_CHANGELOG_LIMIT
        )

    artifact_file.close()
    return cr_pkginfo


def urlpath_sanitize(*args):
    """
    Join an arbitrary number of strings into a /-separated path.

    Replaces uses of urljoin() that don't want/need urljoin's subtle semantics.

    Returns: single string provided arguments separated by single-slashes

    Args:
        Arbitrary list of arguments to be join()ed
    """
    segments = []
    for a in args + ("",):
        stripped = a.strip("/")
        if stripped:
            segments.append(stripped)
    return "/".join(segments)


def get_sha256(file_path):
    """
    Get sha256 of file.

    Args:
        file_path(string): path of file

    Returns:
        String: SHA256 of file

    """
    try:
        with open(file_path, "rb") as file_obj:
            return sha256(file_obj.read()).hexdigest()
    except FileNotFoundError:
        return None


def is_previous_version(version, target_version):
    """
    Compare version with a target version.

    Able to compare versions with integers only. Returns False for non-integer/non-1.2.3 versions.

    Args:
        version(str): version to compare
        target_version(str): version to compare with

    Returns:
        bool: True if versions are the same or if the version is older than the target version.

    """
    # if any of the versions are empty, they can't be compared and
    # a target_version need to be picked
    if version is None or target_version is None:
        return True

    # Handle equals
    if version == target_version:
        return True

    # Handle integers
    if version.isdigit() and target_version.isdigit():
        return int(version) <= int(target_version)

    # Handle 1.2.3
    version_components = version.split(".")
    target_version_components = target_version.split(".")

    if len(version_components) == len(target_version_components):
        for comp_a, comp_b in zip(version_components, target_version_components):
            if comp_a.isdigit() and comp_b.isdigit():
                # if both strings contain numeric information, convert them to ints and compare
                a = int(comp_a)
                b = int(comp_b)
                if a == b:
                    # this 'place' is equal, move to next
                    continue
                else:
                    # not equal, return comparison
                    return a < b
            else:
                # if one of the versions contains non-numeric information we cannot compare
                return False
        else:
            # All places compared, is_prev=>T if last place is equal
            return int(comp_a) == int(comp_b)
    # len wasn't equal
    return False


def parse_time(value):
    """
    Parse datetime values from a string.

    Able to distinguish between timestamp and iso format.

    Args:
        value(str): unformatted time value

    Returns:
        int | datetime | None: formatted time value
    """
    return int(value) if value.isdigit() else parse_datetime(value)


def _get_datapkg_sample_rpm_copy(basedir: str):
    sample_rpm = files("pulp_rpm").joinpath("tests/sample-rpm-0-0.x86_64.rpm")
    copy_rpm = shutil.copy(sample_rpm, basedir)
    return Path(copy_rpm)


class RpmTool:
    """
    A wrapper utility for rpm cli tool.

    Args:
        root: Alternative root directory passed to `rpm --root`
    """

    INVALID_SIGNATURE_ERROR_MSG = "Signature is invalid or pubkey is unreachable"
    UNKNOWN_ERROR_MSG = "Some unknown error occurred"
    UNSIGNED_ERROR_MSG = "The package is not signed"

    def __init__(self, root: t.Optional[Path] = None):
        completed_process = subprocess.run(
            ["which", "rpmsign"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if completed_process.returncode != 0:
            raise RuntimeError("Rpm cli tool is not installed on your system.")

        self.opts = ["--root", str(root.absolute())] if root else []

    @staticmethod
    def get_empty_rpm(basedir: str) -> Path:
        """
        Get an empty rpm package.

        Args:
            basedir: The dir where the rpm will be placed.
        """
        return _get_datapkg_sample_rpm_copy(basedir)

    def import_pubkey_file(self, pubkey_file: str):
        """
        Import public_key file (ascii-armored) into the rpm-tool.

        Args:
            import_pubkey: The public key file in ascii-armored format.
        """
        cmd = ("rpm", *self.opts, "--import", pubkey_file)
        completed_process = subprocess.run(cmd, capture_output=True)
        if completed_process.returncode != 0:
            raise RuntimeError(
                f"Could not import public key into rpm-tool:\n{completed_process.stderr.decode()}"
            )

    def import_pubkey_string(self, pubkey_content: str):
        """
        Import public_key string (ascii-armored) into the rpm-tool.

        Parameters:
            import_pubkey: The public key string in ascii-armored format.
        """
        with tempfile.NamedTemporaryFile() as pubkey_file:
            pubkey_file.write(pubkey_content.encode())
            pubkey_file.flush()
            self.import_pubkey_file(pubkey_file.name)

    def verify_signature(self, rpm_package_file: Path, raises=True):
        """
        Verify that an Rpm Package is signed by some of the imported pubkey.

        Parameters:
            rpm_package_file: Path object to rpm package

        Returns:
            True (if has valid)

        Raises:
            InvalidSignature (for invalid/unsigned package)

        Notes:
            This is based on the command: `rpm --checksig camel-0.1-1.noarch.rpm`
            Which have the following scenarios/outputs:

            - unsigned:
                * returncode: 0
                * output: "camel-0.1-1.noarch.rpm: digests OK"
            - signed, but rpm doesnt have pubkey imported:
                * returncode: 1
                * output: "camel-0.1-1.noarch.rpm: digests SIGNATURES NOT OK"
            - signed and rpm can validate:
                * returncode: 0
                * output: "camel-0.1-1.noarch.rpm: digests signatures OK"
        """
        cmd = ("rpm", *self.opts, "--checksig", str(rpm_package_file.resolve()))
        completed_process = subprocess.run(cmd, capture_output=True)
        stdout = completed_process.stdout.decode()
        stderr = completed_process.stderr.decode()
        output = f"\nstdout: {stdout}\nstderr: {stderr}"
        if completed_process.returncode != 0:
            if "SIGNATURES NOT OK" in stdout:
                raise InvalidSignatureError(f"{RpmTool.INVALID_SIGNATURE_ERROR_MSG}: {output}")
            raise TypeError(f"{RpmTool.UNKNOWN_ERROR_MSG}: {output}")
        elif "signatures" not in output:
            raise InvalidSignatureError(f"{RpmTool.UNSIGNED_ERROR_MSG}: {output}")
        return True



import json
import re
import sys
from gettext import gettext as _

from argparse import RawDescriptionHelpFormatter
from django.core.management import BaseCommand
from django.conf import settings

from pulp_rpm.app.models import RpmRepository
from pulpcore.plugin.util import get_url, extract_pk


def gather_repository_sizes(
    repositories, include_on_demand=False, include_published_metadata=False
):
    """
    Creates a list containing the size report for given repositories.

    Each entry in the list will contain a dict with following minimal fields:
        - name: name of the repository
        - href: href of the repository
        - disk-size: size in bytes of all artifacts stored on disk in the repository

    Each entry can additionally have the optional fields if specified:
        - on-demand-size: approximate size in bytes of all on-demand artifacts in the repository
        - published-metadata-size: size in bytes of all published metadata for the repository

    **Note**: This does not account for the fact that the same artifact can appear in multiple
    repositories without incurring additional disk storage use. User interpretation of these numbers
    for individual repositories should consider totals across multiple repositories in the context
    that artifacts may be shared.
    """
    full_report = []
    for repo in repositories.order_by("name").iterator():
        report = {"name": repo.name, "href": get_url(repo), "disk-size": repo.disk_size}
        if include_on_demand:
            report["on-demand-size"] = repo.on_demand_size
        if include_published_metadata:
            report["published-metadata-size"] = repo.published_metadata_size
        full_report.append(report)

    return full_report


def href_list_handler(value):
    """Common list parsing for a string of hrefs."""
    r = rf"({settings.API_ROOT}(?:[-_a-zA-Z0-9]+/)?api/v3/repositories/[-_a-z]+/[-_a-z]+/[-a-f0-9]+/)"  # noqa: E501
    return re.findall(r, value)


class Command(BaseCommand):
    """Django management command for calculating the storage size of a repository."""

    help = __doc__ + gather_repository_sizes.__doc__

    def add_arguments(self, parser):
        """Set up arguments."""
        parser.add_argument(
            "--repositories",
            type=href_list_handler,
            required=False,
            help=_(
                "List of repository hrefs to generate the report from. Leave blank to include"
                " all repositories."
            ),
        )
        parser.add_argument(
            "--include-on-demand",
            action="store_true",
            help=_("Include the approximate on-demand artifact sizes"),
        )
        parser.add_argument(
            "--include-published-metadata",
            action="store_true",
            help=_("Include the size for the published metadata"),
        )

        parser.formatter_class = RawDescriptionHelpFormatter

    def handle(self, *args, **options):
        """Implement the command."""
        repository_hrefs = options.get("repositories")

        repositories = RpmRepository.objects.all()
        if repository_hrefs:
            repos_ids = [extract_pk(r) for r in repository_hrefs]
            repositories = repositories.filter(pk__in=repos_ids)

        report = gather_repository_sizes(
            repositories,
            include_on_demand=options["include_on_demand"],
            include_published_metadata=options["include_published_metadata"],
        )
        json.dump({"repositories": report}, sys.stdout, indent=4)
        print()



"""Tests to test advisory conflict resolution functionality."""

from pulp_rpm.tests.functional.constants import (
    RPM_ADVISORY_DIFFERENT_PKGLIST_URL,
    RPM_ADVISORY_TEST_ID,
    RPM_UNSIGNED_FIXTURE_URL,
)


def test_two_advisories_same_id_to_repo(
    rpm_repository_api,
    rpm_advisory_api,
    rpm_repository_factory,
    init_and_sync,
    monitor_task,
    delete_orphans_pre,
):
    """
    Test when two different advisories with the same id are added to a repo.

    Should merge the two advisories into a single one.
    """
    # Setup
    repo_rpm_unsigned, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL)
    repo_rpm_advisory_diffpkgs, _ = init_and_sync(url=RPM_ADVISORY_DIFFERENT_PKGLIST_URL)
    advisory_rpm_unsigned_href = (
        rpm_advisory_api.list(
            repository_version=repo_rpm_unsigned.latest_version_href,
            id=RPM_ADVISORY_TEST_ID,
        )
        .results[0]
        .pulp_href
    )
    advisory_rpm_advisory_diffpkgs_href = (
        rpm_advisory_api.list(
            repository_version=repo_rpm_advisory_diffpkgs.latest_version_href,
            id=RPM_ADVISORY_TEST_ID,
        )
        .results[0]
        .pulp_href
    )

    # Test advisory conflicts
    repo = rpm_repository_factory()

    data = {
        "add_content_units": [
            advisory_rpm_unsigned_href,
            advisory_rpm_advisory_diffpkgs_href,
        ]
    }
    response = rpm_repository_api.modify(repo.pulp_href, data)
    monitor_task(response.task)
    a_repo = rpm_repository_api.read(repo.pulp_href)

    duplicated_advisory_list = rpm_advisory_api.list(
        repository_version=a_repo.latest_version_href,
        id=RPM_ADVISORY_TEST_ID,
    ).results
    assert 1 == len(duplicated_advisory_list)



from gettext import gettext as _

from django.conf import settings
from drf_spectacular.utils import extend_schema_serializer
from jsonschema import Draft7Validator
from pulpcore.plugin.models import (
    AsciiArmoredDetachedSigningService,
    Publication,
    Remote,
    Content,
    RepositoryVersion,
)
from pulpcore.plugin.serializers import (
    DetailRelatedField,
    DistributionSerializer,
    PublicationSerializer,
    RelatedField,
    RemoteSerializer,
    RepositorySerializer,
    RepositorySyncURLSerializer,
    ValidateFieldsMixin,
)
from pulpcore.plugin.util import get_domain, resolve_prn
from rest_framework import serializers

from pulp_rpm.app.constants import (
    ALLOWED_CHECKSUM_ERROR_MSG,
    ALLOWED_PUBLISH_CHECKSUM_ERROR_MSG,
    ALLOWED_PUBLISH_CHECKSUMS,
    CHECKSUM_CHOICES,
    COMPRESSION_CHOICES,
    LAYOUT_CHOICES,
    SKIP_TYPES,
    SYNC_POLICY_CHOICES,
)
from pulp_rpm.app.models import (
    RpmDistribution,
    RpmPackageSigningService,
    RpmPublication,
    RpmRemote,
    RpmRepository,
    UlnRemote,
)
from pulp_rpm.app.schema import COPY_CONFIG_SCHEMA
from urllib.parse import urlparse
from textwrap import dedent

# avoid calling into dynaconf many times
ALLOWED_CONTENT_CHECKSUMS = settings.ALLOWED_CONTENT_CHECKSUMS


@extend_schema_serializer(
    deprecate_fields=[
        "metadata_checksum_type",
        "package_checksum_type",
        "gpgcheck",
        "repo_gpgcheck",
        "sqlite_metadata",
    ]
)
class RpmRepositorySerializer(RepositorySerializer):
    """
    Serializer for Rpm Repositories.
    """

    autopublish = serializers.BooleanField(
        help_text=_(
            "Whether to automatically create publications for new repository versions, "
            "and update any distributions pointing to this repository."
        ),
        default=False,
        required=False,
    )
    metadata_signing_service = RelatedField(
        help_text="A reference to an associated signing service.",
        view_name="signing-services-detail",
        queryset=AsciiArmoredDetachedSigningService.objects.all(),
        many=False,
        required=False,
        allow_null=True,
    )
    package_signing_service = RelatedField(
        help_text="A reference to an associated package signing service.",
        view_name="signing-services-detail",
        queryset=RpmPackageSigningService.objects.all(),
        many=False,
        required=False,
        allow_null=True,
    )
    package_signing_fingerprint = serializers.CharField(
        help_text=_(
            "The pubkey V4 fingerprint (160 bits) to be passed to the package signing service."
            "The signing service will use that on signing operations related to this repository."
        ),
        max_length=40,
        required=False,
        allow_blank=True,
        default="",
    )
    retain_package_versions = serializers.IntegerField(
        help_text=_(
            "The number of versions of each package to keep in the repository; "
            "older versions will be purged. The default is '0', which will disable "
            "this feature and keep all versions of each package."
        ),
        min_value=0,
        required=False,
    )
    checksum_type = serializers.ChoiceField(
        help_text=_("The preferred checksum type during repo publish."),
        choices=CHECKSUM_CHOICES,
        required=False,
        allow_null=True,
    )
    metadata_checksum_type = serializers.ChoiceField(
        help_text=_(
            "REMOVED: The checksum type to use for metadata. Not operational since pulp_rpm "
            "3.30.0 release. Use 'checksum_type' instead."
        ),
        choices=CHECKSUM_CHOICES,
        read_only=True,
    )
    package_checksum_type = serializers.ChoiceField(
        help_text=_(
            "REMOVED: The checksum type for packages. Not operational since pulp_rpm 3.30.0 "
            "release. Use 'checksum_type' instead."
        ),
        choices=CHECKSUM_CHOICES,
        required=False,
        allow_null=True,
        read_only=True,
    )
    compression_type = serializers.ChoiceField(
        help_text=_("The compression type to use for metadata files."),
        choices=COMPRESSION_CHOICES,
        required=False,
        allow_null=True,
    )
    layout = serializers.ChoiceField(
        help_text=_("How to layout the packages within the published repository."),
        choices=LAYOUT_CHOICES,
        required=False,
        allow_null=True,
    )
    gpgcheck = serializers.IntegerField(
        help_text=_(
            "REMOVED: An option specifying whether a client should perform a GPG signature "
            "check on packages. Not operational since pulp_rpm 3.30.0 release. "
            "Set these values using 'repo_config' instead."
        ),
        max_value=1,
        min_value=0,
        read_only=True,
    )
    repo_gpgcheck = serializers.IntegerField(
        help_text=_(
            "REMOVED: An option specifying whether a client should perform a GPG signature "
            "check on the repodata. Not operational since pulp_rpm 3.30.0 release. "
            "Set these values using 'repo_config' instead."
        ),
        max_value=1,
        min_value=0,
        read_only=True,
    )
    sqlite_metadata = serializers.BooleanField(
        help_text=_(
            "REMOVED: An option specifying whether Pulp should generate SQLite metadata. "
            "Not operation since pulp_rpm 3.25.0 release"
        ),
        default=False,
        read_only=True,
    )
    repo_config = serializers.JSONField(
        required=False,
        help_text=_(
            "A JSON document describing the config.repo file Pulp should generate for this repo"
        ),
    )

    def to_representation(self, instance):
        data = super().to_representation(instance)
        # Import workflow may cause these fields to be stored as "" in the database
        # This ensure the correct type of  None | Enum in the response
        for field in (
            "checksum_type",
            "metadata_checksum_type",
            "package_checksum_type",
            "compression_type",
            "layout",
        ):
            field_data = data.get(field)
            if field_data == "":
                data[field] = None
        # The current API field definition expects empty string for nullable values,
        # but some migration paths can set an empty string to none in the database.
        if "package_signing_fingerprint" in data and data["package_signing_fingerprint"] is None:
            data["package_signing_fingerprint"] = ""
        return data

    def validate(self, data):
        """Validate data."""
        if checksum_type := data.get("checksum_type"):
            if checksum_type not in ALLOWED_CONTENT_CHECKSUMS:
                raise serializers.ValidationError({"checksum_type": _(ALLOWED_CHECKSUM_ERROR_MSG)})

            if checksum_type not in ALLOWED_PUBLISH_CHECKSUMS:
                raise serializers.ValidationError(
                    {"checksum_type": _(ALLOWED_PUBLISH_CHECKSUM_ERROR_MSG)}
                )

        validated_data = super().validate(data)
        return validated_data

    class Meta:
        fields = RepositorySerializer.Meta.fields + (
            "autopublish",
            "metadata_signing_service",
            "package_signing_service",
            "package_signing_fingerprint",
            "retain_package_versions",
            "checksum_type",
            "metadata_checksum_type",
            "package_checksum_type",
            "gpgcheck",
            "repo_gpgcheck",
            "sqlite_metadata",
            "repo_config",
            "compression_type",
            "layout",
        )
        model = RpmRepository


class RpmBaseRemoteSerializer(RemoteSerializer):
    """
    A common base serializer for multiple RPM based remotes.
    """

    policy = serializers.ChoiceField(
        help_text=_(
            "The policy to use when downloading content. The possible values include: "
            "'immediate', 'on_demand', and 'streamed'. 'immediate' is the default."
        ),
        choices=Remote.POLICY_CHOICES,
        default=Remote.IMMEDIATE,
    )


class RpmRemoteSerializer(RpmBaseRemoteSerializer):
    """
    A Serializer for RpmRemote.
    """

    sles_auth_token = serializers.CharField(
        help_text=_("Authentication token for SLES repositories."),
        required=False,
        allow_null=True,
    )

    def validate_url(self, value):
        ALLOWED = ("http", "https", "file")
        protocol = urlparse(value).scheme
        if protocol not in ALLOWED:
            raise serializers.ValidationError(
                f"The url {repr(value)} is not valid. It must start with: {ALLOWED}."
            )
        return value

    class Meta:
        fields = RemoteSerializer.Meta.fields + ("sles_auth_token",)
        model = RpmRemote


class UlnRemoteSerializer(RpmBaseRemoteSerializer):
    """
    A Serializer for UlnRemote.
    """

    username = serializers.CharField(
        help_text=_("Your ULN account username."),
        required=True,
        write_only=True,
    )
    password = serializers.CharField(
        help_text=_("Your ULN account password."),
        required=True,
        write_only=True,
        style={"input_type": "password"},
    )

    url = serializers.CharField(
        help_text=_(
            "The ULN repo URL of the remote content source."
            '"This is "uln://" followed by the channel name. E.g.: "uln://ol7_x86_64_oracle"'
        ),
        required=True,
    )

    uln_server_base_url = serializers.CharField(
        help_text=_(
            "Base URL of the ULN server. If the uln_server_base_url is not provided pulp_rpm will"
            "use the contents of the DEFAULT_ULN_SERVER_BASE_URL setting instead."
        ),
        required=False,
        allow_null=True,
    )

    def validate_url(self, value):
        ALLOWED = ("uln",)
        protocol = urlparse(value).scheme
        if protocol not in ALLOWED:
            raise serializers.ValidationError(
                f"The url {repr(value)} is not valid. It must start with: {ALLOWED}."
            )
        return value

    class Meta:
        fields = RemoteSerializer.Meta.fields + ("uln_server_base_url",)
        model = UlnRemote


@extend_schema_serializer(
    deprecate_fields=[
        "metadata_checksum_type",
        "package_checksum_type",
        "gpgcheck",
        "repo_gpgcheck",
        "sqlite_metadata",
    ]
)
class RpmPublicationSerializer(PublicationSerializer):
    """
    A Serializer for RpmPublication.
    """

    metadata_checksum_type = serializers.ChoiceField(
        help_text=_(
            "REMOVED: The checksum type for metadata. Not operational since pulp_rpm 3.30.0 "
            "release. Use 'checksum_type' instead."
        ),
        choices=CHECKSUM_CHOICES,
        read_only=True,
    )
    package_checksum_type = serializers.ChoiceField(
        help_text=_(
            "REMOVED: The checksum type for packages. Not operational since pulp_rpm 3.30.0 "
            "release. Use 'checksum_type' instead."
        ),
        choices=CHECKSUM_CHOICES,
        read_only=True,
    )
    checkpoint = serializers.BooleanField(required=False)
    checksum_type = serializers.ChoiceField(
        help_text=_("The preferred checksum type used during repo publishes."),
        choices=CHECKSUM_CHOICES,
        required=False,
    )
    compression_type = serializers.ChoiceField(
        help_text=_("The compression type to use for metadata files."),
        choices=COMPRESSION_CHOICES,
        required=False,
    )
    layout = serializers.ChoiceField(
        help_text=_("How to layout the packages within the published repository."),
        choices=LAYOUT_CHOICES,
        required=False,
        allow_null=True,
    )
    gpgcheck = serializers.IntegerField(
        help_text=_(
            "REMOVED: An option specifying whether a client should perform "
            "a GPG signature check on packages. Not operational since pulp_rpm 3.30.0 release. "
            "Set these values using 'repo_config' instead."
        ),
        max_value=1,
        min_value=0,
        read_only=True,
    )
    repo_gpgcheck = serializers.IntegerField(
        help_text=_(
            "REMOVED: An option specifying whether a client should perform "
            "a GPG signature check on the repodata. Not operational since pulp_rpm 3.30.0 release. "
            "Set these values using 'repo_config' instead."
        ),
        max_value=1,
        min_value=0,
        read_only=True,
    )
    sqlite_metadata = serializers.BooleanField(
        help_text=_(
            "REMOVED: An option specifying whether Pulp should generate SQLite metadata. "
            "Not operational since pulp_rpm 3.25.0 release"
        ),
        default=False,
        read_only=True,
    )
    repo_config = serializers.JSONField(
        required=False,
        help_text=_(
            "A JSON document describing the config.repo file Pulp should generate for this repo"
        ),
    )

    def validate(self, data):
        """Validate data."""
        if checksum_type := data.get("checksum_type"):
            if checksum_type not in ALLOWED_CONTENT_CHECKSUMS:
                raise serializers.ValidationError(ALLOWED_CHECKSUM_ERROR_MSG)

            if checksum_type not in ALLOWED_PUBLISH_CHECKSUMS:
                raise serializers.ValidationError(ALLOWED_PUBLISH_CHECKSUM_ERROR_MSG)

        validated_data = super().validate(data)
        return validated_data

    class Meta:
        fields = PublicationSerializer.Meta.fields + (
            "checkpoint",
            "checksum_type",
            "metadata_checksum_type",
            "package_checksum_type",
            "gpgcheck",
            "repo_gpgcheck",
            "sqlite_metadata",
            "repo_config",
            "compression_type",
            "layout",
        )
        model = RpmPublication


class RpmDistributionSerializer(DistributionSerializer):
    """
    Serializer for RPM Distributions.
    """

    publication = DetailRelatedField(
        required=False,
        help_text=_("Publication to be served"),
        view_name_pattern=r"publications(-.*/.*)?-detail",
        queryset=Publication.objects.exclude(complete=False),
        allow_null=True,
    )
    generate_repo_config = serializers.BooleanField(
        default=False,
        required=False,
        help_text=_("An option specifying whether Pulp should generate *.repo files."),
    )
    checkpoint = serializers.BooleanField(required=False)

    class Meta:
        fields = DistributionSerializer.Meta.fields + (
            "publication",
            "generate_repo_config",
            "checkpoint",
        )
        model = RpmDistribution


class RpmRepositorySyncURLSerializer(RepositorySyncURLSerializer):
    """
    Serializer for RPM Sync.
    """

    mirror = serializers.BooleanField(
        required=False,
        allow_null=True,
        help_text=_(
            "DEPRECATED: If ``True``, ``sync_policy`` will default to 'mirror_complete' "
            "instead of 'additive'."
        ),
    )
    sync_policy = serializers.ChoiceField(
        help_text=_(
            "Options: 'additive', 'mirror_complete', 'mirror_content_only'. Default: 'additive'. "
            "Modifies how the sync is performed. 'mirror_complete' will clone the original "
            "metadata and create an automatic publication from it, but comes with some "
            "limitations and does not work for certain repositories. 'mirror_content_only' will "
            "change the repository contents to match the remote but the metadata will be "
            "regenerated and will not be bit-for-bit identical. 'additive' will retain the "
            "existing contents of the repository and add the contents of the repository being "
            "synced."
        ),
        choices=SYNC_POLICY_CHOICES,
        required=False,
        allow_null=True,
    )
    skip_types = serializers.ListField(
        help_text=_("List of content types to skip during sync."),
        required=False,
        default=[],
        child=serializers.ChoiceField([(skip_type, skip_type) for skip_type in SKIP_TYPES]),
    )
    optimize = serializers.BooleanField(
        help_text=_("Whether or not to optimize sync."), required=False, default=True
    )

    def validate(self, data):
        """
        Validate sync parameters.
        """
        data = super().validate(data)

        if "mirror" in data and "sync_policy" in data:
            raise serializers.ValidationError(
                _(
                    "Cannot use 'mirror' and 'sync_policy' options simultaneously. The 'mirror' "
                    "option is deprecated, please use 'sync_policy' only."
                )
            )

        return data


class CopySerializer(ValidateFieldsMixin, serializers.Serializer):
    """
    A serializer for Content Copy API.
    """

    config = serializers.JSONField(
        help_text=_(
            dedent(
                """\
        Content to be copied into the given destinations from the given sources.

        Its a list of dictionaries with the following available fields:

        ```json
        [
          {
            "source_repo_version": <RepositoryVersion [pulp_href|prn]>,
            "dest_repo": <RpmRepository [pulp_href|prn]>,
            "dest_base_version": <int>,
            "content": [<Content [pulp_href|prn]>, ...]
          },
          ...
        ]
        ```

        If domains are enabled, the refered pulp objects must be part of the current domain.

        For usage examples, refer to the advanced copy guide:
        <https://pulpproject.org/pulp_rpm/docs/user/guides/modify/#advanced-copy-workflow>
        """
            )
        ),
    )

    dependency_solving = serializers.BooleanField(
        help_text=_("Also copy dependencies of the content being copied."), default=True
    )

    def validate(self, data):
        """
        Validate that the Serializer contains valid data.

        Make sure the config-JSON matches the config-schema.
        Check for cross-domain references (if domain-enabled).
        """

        def raise_validation(field, domain, id=""):
            id = f"\n{id}" if id else ""
            raise serializers.ValidationError(
                _("The field {} contains object(s) not in {} domain.{}".format(field, domain, id))
            )

        def parse_reference(ref) -> tuple[str, str, bool]:
            """Extract info from prn/href to enable checking domains.

            This is used for:
            1. In case of HREFS, avoid expensive extract_pk(href) to get pks.
            2. HREF and PRNs have different information hardcoded available.
               E.g: RepositoryVerseion HREF has its Repository pk; PRNs have the RepoVer pk.

            Returns a tuple with (pk, class_name, is_prn)
            """
            if ref.startswith("prn:"):
                ref_class, pk = resolve_prn(ref)
                return (pk, ref_class, True)
            # content:    ${BASE}/content/rpm/packages/${UUID}/
            # repository: ${BASE}/repositories/rpm/rpm/${UUID}/
            # repover:    ${BASE}/repositories/rpm/rpm/${UUID}/versions/0/
            url = urlparse(ref).path.strip("/").split("/")
            ref_class = RpmRepository if "/repositories/" in ref else Content
            is_repover_href = url[-1].isdigit() and url[-2] == "versions"
            uuid = url[-3] if is_repover_href else url[-1]
            if len(uuid) < 32:
                raise serializers.ValidationError(
                    _("The href path should end with a uuid pk: {}".format(ref))
                )
            return (uuid, ref_class, False)

        def check_domain(entry, name, curr_domain):
            """Check domain for RpmRepository and RepositoryVersion objects."""
            href_or_prn = entry[name]
            resource_pk, ref_class, is_prn = parse_reference(href_or_prn)
            try:
                if ref_class is RepositoryVersion and is_prn:
                    resource_domain_pk = (
                        RepositoryVersion.objects.select_related("repository")
                        .values("repository__pulp_domain")
                        .get(pk=resource_pk)["repository__pulp_domain"]
                    )
                elif ref_class is RpmRepository:
                    resource_domain_pk = RpmRepository.objects.values("pulp_domain").get(
                        pk=resource_pk
                    )["pulp_domain"]
                else:
                    raise serializers.ValidationError(
                        _(
                            "Expected RpmRepository or RepositoryVersion ref_class. "
                            "Got {} from {}.".format(ref_class, href_or_prn)
                        )
                    )
            except RepositoryVersion.DoesNotExit as e:
                raise serializers.ValidationError from e
            except RpmRepository.DoesNotExit as e:
                raise serializers.ValidationError from e

            if resource_domain_pk != curr_domain.pk:
                raise_validation(name, curr_domain.name, resource_domain_pk)

        def check_cross_domain_config(cfg):
            """Check that all config-elts are in 'our' domain."""
            # copy-cfg is a list of dictionaries.
            # source_repo_version and dest_repo are required fields.
            # Insure curr-domain exists in src/dest/dest_base_version/content-list hrefs
            curr_domain = get_domain()
            for entry in cfg:
                # Check required fields individually
                check_domain(entry, "source_repo_version", curr_domain)
                check_domain(entry, "dest_repo", curr_domain)

                # Check content generically to avoid timeout of multiple calls
                content_list = entry.get("content", None)
                if content_list:
                    content_list = [parse_reference(v)[0] for v in content_list]
                    distinct = (
                        Content.objects.filter(pk__in=content_list).values("pulp_domain").distinct()
                    )
                    domain_ok = (
                        len(distinct) == 1 and distinct.first()["pulp_domain"] == curr_domain.pk
                    )
                    if not domain_ok:
                        raise_validation("content", curr_domain.name)

        super().validate(data)
        if "config" in data:
            # Make sure config is valid JSON
            validator = Draft7Validator(COPY_CONFIG_SCHEMA)

            err = []
            for error in sorted(validator.iter_errors(data["config"]), key=str):
                err.append(error.message)
            if err:
                raise serializers.ValidationError(
                    _("Provided copy criteria is invalid:'{}'".format(err))
                )

            if settings.DOMAIN_ENABLED:
                check_cross_domain_config(data["config"])

        return data



from drf_spectacular.utils import extend_schema
from rest_framework import viewsets

from pulpcore.plugin.models import PulpTemporaryFile
from pulpcore.plugin.tasking import dispatch
from pulpcore.plugin.serializers import (
    AsyncOperationResponseSerializer,
)
from pulpcore.plugin.viewsets import (
    OperationPostponedResponse,
    ReadOnlyContentViewSet,
)

from pulp_rpm.app import tasks
from pulp_rpm.app.models import (
    PackageCategory,
    PackageEnvironment,
    PackageGroup,
    PackageLangpacks,
)
from pulp_rpm.app.serializers import (
    CompsXmlSerializer,
    PackageCategorySerializer,
    PackageEnvironmentSerializer,
    PackageGroupSerializer,
    PackageLangpacksSerializer,
)


class CompsXmlViewSet(viewsets.ViewSet):
    """
    ViewSet for comps.xml Upload.
    """

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_required_repo_perms_on_upload:rpm.modify_content_rpmrepository",
                    "has_required_repo_perms_on_upload:rpm.view_rpmrepository",
                ],
            },
        ],
    }

    @extend_schema(
        description="Trigger an asynchronous task to upload a comps.xml file.",
        summary="Upload comps.xml",
        operation_id="rpm_comps_upload",
        request=CompsXmlSerializer,
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request):
        """Upload a comps.xml file and create Content from it."""
        serializer = CompsXmlSerializer(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)

        # Store TemporaryUpload as a file we can find/use from our task
        task_payload = {k: v for k, v in request.data.items()}
        file_content = task_payload.pop("file", None)
        temp_file = PulpTemporaryFile.init_and_validate(file_content)
        temp_file.save()

        # Lock destination-repo if we are given one so two uploads can't collide
        repository = serializer.validated_data.get("repository", None)
        repo_pk = str(repository.pk) if repository else None
        replace = serializer.validated_data.get("replace", False)

        # Kick off task to Do the Deed
        task = dispatch(
            tasks.upload_comps,
            exclusive_resources=[repository] if repository else [],
            args=([str(temp_file.pk), repo_pk, replace]),
            kwargs={},
        )
        return OperationPostponedResponse(task, request)


class PackageGroupViewSet(ReadOnlyContentViewSet):
    """
    PackageGroup ViewSet.
    """

    endpoint_name = "packagegroups"
    queryset = PackageGroup.objects.all()
    serializer_class = PackageGroupSerializer

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }


class PackageCategoryViewSet(ReadOnlyContentViewSet):
    """
    PackageCategory ViewSet.
    """

    endpoint_name = "packagecategories"
    queryset = PackageCategory.objects.all()
    serializer_class = PackageCategorySerializer

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }


class PackageEnvironmentViewSet(ReadOnlyContentViewSet):
    """
    PackageEnvironment ViewSet.
    """

    endpoint_name = "packageenvironments"
    queryset = PackageEnvironment.objects.all()
    serializer_class = PackageEnvironmentSerializer

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }


class PackageLangpacksViewSet(ReadOnlyContentViewSet):
    """
    PackageLangpacks ViewSet.
    """

    endpoint_name = "packagelangpacks"
    queryset = PackageLangpacks.objects.all()
    serializer_class = PackageLangpacksSerializer

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }



# Generated by Django 2.2.10 on 2020-02-19 14:43


from django.db import models, migrations, transaction


def replace_nonestring_with_null(apps, schema_editor):
    with transaction.atomic():
        UpdateRecord = apps.get_model('rpm', 'UpdateRecord')

        recs = list(UpdateRecord.objects.filter(updated_date__in=('None', '')).only('updated_date'))
        for rec in recs:
            rec.updated_date = None

        UpdateRecord.objects.bulk_update(recs, ['updated_date'])


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0029_rpmpublication_sqlite_metadata'),
    ]

    operations = [
        migrations.AlterField(
            model_name='updaterecord',
            name='updated_date',
            field=models.TextField(null=True),
        ),
        # The string "None" was stored in the database in place of "" or null
        migrations.RunPython(replace_nonestring_with_null),
    ]
